#!/usr/bin/env python
# coding=utf-8
# Filename: io.py

"""
IO Code for OrcaSong.
"""

import os
import errno
import toml
import warnings


def load_config(config_filepath):
    """
    Loads the config from a .toml file.

    Parameters
    ----------
    config_filepath : str
        Full filepath to a config file. An example can be found in orcasong/default_config.toml

    Returns
    -------
    config : dict
        Dictionary that contains all configuration options of the make_nn_images function.
        An explanation of the config parameters can be found in orcasong/default_config.toml.

    """
    config = toml.load(config_filepath)
    print('Loaded the config file from ' + os.path.abspath(config_filepath))

    return config


def check_user_input(fname, detx_filepath, config):
    """
    Sanity check of the user input.

    Parameters
    ----------
    fname : str
        Full filepath to the input .h5 file.
    detx_filepath : str
        Full filepath to the .detx geometry file that belongs to the fname.
    config : dict
        Dictionary that contains all configuration options of the make_nn_images function.
        An explanation of the config parameters can be found in orcasong/default_config.toml.

    """
    #---- Checks input types ----#

    # Check for options with a single, non-boolean element
    number_args = {'do2d_plots_n': int,  'data_cut_e_low': float, 'data_cut_e_high': float,
                   'data_cut_throw_away': float, 'prod_ident': int}

    for key in number_args:
        expected_arg_type = number_args[key]
        parsed_arg = config[key]

        if parsed_arg in [None, 'None']: # we don't want to check args when there has been no user input
            continue

        if type(parsed_arg) != expected_arg_type:
            try:
                map(expected_arg_type, parsed_arg)
            except ValueError:
                raise TypeError('The argument option ', key, ' only accepts ', str(expected_arg_type),
                                ' values as an input.')

    # Checks the n_bins tuple input
    for dim in config['n_bins']:
        if type(dim) != int:
            raise TypeError('The argument option n_bins only accepts integer values as an input!'
                            ' Your values have the type ' + str(type(dim)))

    # ---- Checks input types ----#

    # ---- Check other things ----#

    if not os.path.isfile(fname):
        raise IOError('The file -' + fname+ '- does not exist.')

    if not os.path.isfile(detx_filepath):
        raise IOError('The file -' + detx_filepath + '- does not exist.')

    if all(do_nd == False for do_nd in [config['do2d'], config['do3d'],config['do4d']]):
        raise ValueError('At least one of do2d, do3d or do4d options must be set to True.')

    if config['do2d'] == False and config['do2d_plots'] == True:
        raise ValueError('The 2D pdf images cannot be created if do2d=False!')

    if config['do2d_plots'] == True and config['do2d_plots_n'] > 100:
        warnings.warn('You declared do2d_pdf=(True, int) with int > 100. This will take more than two minutes.'
                      'Do you really want to create pdfs images for so many events?')


def make_output_dirs(output_dirpath, do2d, do3d, do4d):
    """
    Function that creates all output directories if they don't exist already.

    Parameters
    ----------
    output_dirpath : str
        Full path to the directory, where the orcasong output should be stored.
    do2d : bool
        Declares if 2D histograms, are to be created.
    do3d : bool
        Declares if 3D histograms are to be created.
    do4d : tuple(bool, str)
        Tuple that declares if 4D histograms should be created [0] and if yes, what should be used as the 4th dim after xyz.
        Currently, only 'time' and 'channel_id' are available.

    """
    if do2d:
        projections = ['xy', 'xz', 'yz', 'xt', 'yt', 'zt']
        for proj in projections:
            if not os.path.exists(output_dirpath + '/orcasong_output/4dTo2d/' + proj):
                try:
                    os.makedirs(output_dirpath + '/orcasong_output/4dTo2d/' + proj)
                except OSError as e:
                    if e.errno != errno.EEXIST:
                        raise

    if do3d:
        projections = ['xyz', 'xyt', 'xzt', 'yzt', 'rzt']
        for proj in projections:
            if not os.path.exists(output_dirpath + '/orcasong_output/4dTo3d/' + proj):
                try:
                    os.makedirs(output_dirpath + '/orcasong_output/4dTo3d/' + proj)
                except OSError as e:
                    if e.errno != errno.EEXIST:
                        raise

    if do4d[0]:
        proj = 'xyzt' if not do4d[1] == 'channel_id' else 'xyzc'
        if not os.path.exists(output_dirpath + '/orcasong_output/4dTo4d/' + proj):
            try:
                os.makedirs(output_dirpath + '/orcasong_output/4dTo4d/' + proj)
            except OSError as e:
                if e.errno != errno.EEXIST:
                    raise


import os
from abc import abstractmethod
import h5py
import km3pipe as kp
import km3modules as km

import orcasong
import orcasong.modules as modules
import orcasong.plotting.plot_binstats as plot_binstats


__author__ = 'Stefan Reck'


class BaseProcessor:
    """
    Preprocess km3net/antares events for neural networks.

    This serves as a baseclass, which handles things like reading
    events, calibrating, generating labels and saving the output.

    Parameters
    ----------
    extractor : function, optional
        Function that extracts desired info from a blob, which is then
        stored as the "y" datafield in the .h5 file.
        The function takes the km3pipe blob as an input, and returns
        a dict mapping str to floats.
        Examples can be found in orcasong.extractors.
    det_file : str, optional
        Path to a .detx detector geometry file, which can be used to
        calibrate the hits.
    correct_mc_time : bool
        Converts MC hit times to JTE times.
    center_time : bool
        Subtract time of first triggered hit from all hit times. Will
        also be done for McHits if they are in the blob [default: True].
    correct_timeslew : bool
        If true, the time slewing of hits depending on their tot
        will be corrected.
    add_t0 : bool
        If true, add t0 to the time of hits and mchits. If using a
        det_file, this will already have been done automatically
        [default: False].
    event_skipper : func, optional
        Function that takes the blob as an input, and returns a bool.
        If the bool is true, the blob will be skipped.
        This is placed after the binning and mc_info extractor.
    chunksize : int
        Chunksize (along axis_0) used for saving the output
        to a .h5 file [default: 32].
    keep_event_info : bool
        If True, will keep the "event_info" table [default: False].
    overwrite : bool
        If True, overwrite the output file if it exists already.
        If False, throw an error instead.
    mc_info_to_float64 : bool
        Convert everything in the mcinfo array to float 64 (Default: True).
        Hint: int dtypes can not store nan!

    Attributes
    ----------
    n_statusbar : int or None
        Print a statusbar every n blobs.
    n_memory_observer : int or None
        Print memory usage every n blobs.
    complib : str
        Compression library used for saving the output to a .h5 file.
        All PyTables compression filters are available, e.g. 'zlib',
        'lzf', 'blosc', ... .
    complevel : int
        Compression level for the compression filter that is used for
        saving the output to a .h5 file.
    flush_frequency : int
        After how many events the accumulated output should be flushed to
        the harddisk.
        A larger value leads to a faster orcasong execution,
        but it increases the RAM usage as well.
    seed : int, optional
        Makes all random (numpy) actions reproducable. Set at the start of
        each pipeline.

    """
    def __init__(self, extractor=None,
                 det_file=None,
                 correct_mc_time=True,
                 center_time=True,
                 add_t0=False,
                 correct_timeslew=True,
                 event_skipper=None,
                 chunksize=32,
                 keep_event_info=False,
                 overwrite=True,
                 mc_info_to_float64=True):
        self.extractor = extractor
        self.det_file = det_file
        self.correct_mc_time = correct_mc_time
        self.center_time = center_time
        self.add_t0 = add_t0
        self.correct_timeslew = correct_timeslew
        self.event_skipper = event_skipper
        self.chunksize = chunksize
        self.keep_event_info = keep_event_info
        self.overwrite = overwrite
        self.mc_info_to_float64 = mc_info_to_float64

        self.n_statusbar = 1000
        self.n_memory_observer = 1000
        self.complib = 'zlib'
        self.complevel = 1
        self.flush_frequency = 1000
        self.seed = 42

    def run(self, infile, outfile=None):
        """
        Process the events from the infile, and save them to the outfile.

        Parameters
        ----------
        infile : str
            Path to the input file.
        outfile : str, optional
            Path to the output file (will be created). If none is given,
            will auto generate the name and save it in the cwd.

        """
        if outfile is None:
            outfile = os.path.join(os.getcwd(), "{}_dl.h5".format(
                os.path.splitext(os.path.basename(infile))[0]))
        if not self.overwrite:
            if os.path.isfile(outfile):
                raise FileExistsError(f"File exists: {outfile}")
        if self.seed:
            km.GlobalRandomState(seed=self.seed)
        pipe = self.build_pipe(infile, outfile)
        summary = pipe.drain()
        with h5py.File(outfile, "a") as f:
            self.finish_file(f, summary)

    def run_multi(self, infiles, outfolder):
        """
        Process multiple files into their own output files each.
        The output file names will be generated automatically.

        Parameters
        ----------
        infiles : List
            The path to infiles as str.
        outfolder : str
            The output folder to place them in.

        """
        outfiles = []
        for infile in infiles:
            outfile = os.path.join(
                outfolder,
                f"{os.path.splitext(os.path.basename(infile))[0]}_dl.h5")
            outfiles.append(outfile)
            self.run(infile, outfile)
        return outfiles

    def build_pipe(self, infile, outfile, timeit=True):
        """ Initialize and connect the modules from the different stages. """
        components = [
            *self.get_cmpts_pre(infile=infile),
            *self.get_cmpts_main(),
            *self.get_cmpts_post(outfile=outfile),
        ]
        pipe = kp.Pipeline(timeit=timeit)
        if self.n_statusbar is not None:
            pipe.attach(km.common.StatusBar, every=self.n_statusbar)
        if self.n_memory_observer is not None:
            pipe.attach(km.common.MemoryObserver, every=self.n_memory_observer)
        for cmpt, kwargs in components:
            pipe.attach(cmpt, **kwargs)
        return pipe

    def get_cmpts_pre(self, infile):
        """ Modules that read and calibrate the events. """
        cmpts = [(kp.io.hdf5.HDF5Pump, {"filename": infile})]

        if self.det_file:
            cmpts.append((modules.DetApplier, {
                "det_file": self.det_file,
                "correct_timeslew": self.correct_timeslew}))
        if self.correct_mc_time:
            cmpts.append((km.mc.MCTimeCorrector, {}))

        if any((self.center_time, self.add_t0)):
            cmpts.append((modules.TimePreproc, {
                "add_t0": self.add_t0,
                "center_time": self.center_time}))
        return cmpts

    @abstractmethod
    def get_cmpts_main(self):
        """  Produce and store the samples as 'samples' in the blob. """
        raise NotImplementedError

    def get_cmpts_post(self, outfile):
        """ Modules that postproc and save the events. """
        cmpts = []
        if self.extractor is not None:
            cmpts.append((modules.McInfoMaker, {
                "extractor": self.extractor,
                "to_float64": self.mc_info_to_float64,
                "store_as": "mc_info"}))

        if self.event_skipper is not None:
            cmpts.append((modules.EventSkipper, {
                "event_skipper": self.event_skipper}))

        keys_keep = ['samples', 'mc_info', "header", "raw_header"]
        if self.keep_event_info:
            keys_keep.append('EventInfo')
        cmpts.append((km.common.Keep, {"keys": keys_keep}))

        cmpts.append((kp.io.HDF5Sink, {
            "filename": outfile,
            "complib": self.complib,
            "complevel": self.complevel,
            "chunksize": self.chunksize,
            "flush_frequency": self.flush_frequency}))
        return cmpts

    def finish_file(self, f, summary):
        """
        Work with the output file after the pipe has finished.

        Parameters
        ----------
        f : h5py.File
            The opened output file.
        summary : km3pipe.Blob
            The output from pipe.drain().

        """
        # Add current orcasong version to h5 file
        f.attrs.create("orcasong", orcasong.__version__, dtype="S6")


class FileBinner(BaseProcessor):
    """
    For making binned images and mc_infos, which can be used for conv nets.

    Can also add statistics of the binning to the h5 files, which can
    be plotted to show the distribution of hits among the bins and how
    many hits were cut off.

    Parameters
    ----------
    bin_edges_list : List
        List with the names of the fields to bin, and the respective bin
        edges, including the left- and right-most bin edge.
        Example: For 10 bins in the z direction, and 100 bins in time:
            bin_edges_list = [
                ["pos_z", np.linspace(0, 10, 11)],
                ["time", np.linspace(-50, 550, 101)],
            ]
        Some examples can be found in orcasong.bin_edges.
    add_bin_stats : bool
        Add statistics of the binning to the output file. They can be
        plotted with util/bin_stats_plot.py [default: True].
    hit_weights : str, optional
        Use blob["Hits"][hit_weights] as weights for samples in histogram.
    kwargs
        Options of the BaseProcessor.

    """
    def __init__(self, bin_edges_list,
                 add_bin_stats=True,
                 hit_weights=None,
                 **kwargs):
        self.bin_edges_list = bin_edges_list
        self.add_bin_stats = add_bin_stats
        self.hit_weights = hit_weights
        super().__init__(**kwargs)

    def get_cmpts_main(self):
        """ Generate nD images. """
        cmpts = []
        if self.add_bin_stats:
            cmpts.append((modules.BinningStatsMaker, {
                "bin_edges_list": self.bin_edges_list}))
        cmpts.append((modules.ImageMaker, {
            "bin_edges_list": self.bin_edges_list,
            "hit_weights": self.hit_weights}))
        return cmpts

    def finish_file(self, f, summary):
        super().finish_file(f, summary)
        if self.add_bin_stats:
            plot_binstats.add_hists_to_h5file(summary["BinningStatsMaker"], f)

    def run_multi(self, infiles, outfolder, save_plot=False):
        """
        Bin multiple files into their own output files each.
        The output file names will be generated automatically.

        Parameters
        ----------
        infiles : List
            The path to infiles as str.
        outfolder : str
            The output folder to place them in.
        save_plot : bool
            Save the binning hists as a pdf. Only possible if add_bin_stats
            is True.

        """
        if save_plot and not self.add_bin_stats:
            raise ValueError("Can not make plot when add_bin_stats is False")

        name, shape = self.get_names_and_shape()
        print("Generating {} images with shape {}".format(name, shape))

        outfiles = super().run_multi(infiles=infiles, outfolder=outfolder)

        if save_plot:
            plot_binstats.plot_hist_of_files(
                files=outfiles, save_as=outfolder+"binning_hist.pdf")
        return outfiles

    def get_names_and_shape(self):
        """
        Get names and shape of the resulting x data,
        e.g. (pos_z, time), (18, 50).
        """
        names, shape = [], []
        for bin_name, bin_edges in self.bin_edges_list:
            names.append(bin_name)
            shape.append(len(bin_edges) - 1)
        return tuple(names), tuple(shape)

    def __repr__(self):
        return "<FileBinner: {} {}>".format(*self.get_names_and_shape())


class FileGraph(BaseProcessor):
    """
    Turn km3 events to graph data.

    The resulting file will have a dataset "x" of shape
    (?, max_n_hits, len(hit_infos) + 1).
    The column names of the last axis (i.e. hit_infos) are saved
    as attributes of the dataset (f["x"].attrs).
    The last column will always be called 'is_valid', and its 0 if
    the entry is padded, and 1 otherwise.

    Parameters
    ----------
    max_n_hits : int
        Maximum number of hits that gets saved per event. If an event has
        more, some will get cut randomly!
    time_window : tuple, optional
        Two ints (start, end). Hits outside of this time window will be cut
        away (based on 'Hits/time'). Default: Keep all hits.
    hit_infos : tuple, optional
        Which entries in the '/Hits' Table will be kept. E.g. pos_x, time, ...
        Default: Keep all entries.
    kwargs
        Options of the BaseProcessor.

    """
    def __init__(self, max_n_hits,
                 time_window=None,
                 hit_infos=None,
                 **kwargs):
        self.max_n_hits = max_n_hits
        self.time_window = time_window
        self.hit_infos = hit_infos
        super().__init__(**kwargs)

    def get_cmpts_main(self):
        return [((modules.PointMaker, {
            "max_n_hits": self.max_n_hits,
            "time_window": self.time_window,
            "hit_infos": self.hit_infos,
            "dset_n_hits": "EventInfo"}))]

    def finish_file(self, f, summary):
        super().finish_file(f, summary)
        for i, hit_info in enumerate(summary["PointMaker"]["hit_infos"]):
            f["x"].attrs.create(f"hit_info_{i}", hit_info)



from unittest import TestCase
import os
import h5py
import numpy as np
import orcasong.tools.postproc as postproc

__author__ = 'Stefan Reck'

test_dir = os.path.dirname(os.path.realpath(__file__))
MUPAGE_FILE = os.path.join(test_dir, "data", "mupage.root.h5")


class TestPostproc(TestCase):
    def setUp(self):
        self.output_file = "temp_output.h5"

    def tearDown(self):
        if os.path.exists(self.output_file):
            os.remove(self.output_file)

    def test_shuffle(self):
        postproc.postproc_file(
            input_file=MUPAGE_FILE,
            output_file=self.output_file,
            shuffle=True,
            event_skipper=None,
            delete=False,
            seed=13,
        )

        with h5py.File(self.output_file, "r") as f:
            np.testing.assert_equal(f["event_info"]["event_id"], np.array([1, 0, 2]))
            self.assertTrue("origin" in f.attrs.keys())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Contains functions to shuffles .h5 files.

Can only be used for files where each dataset has the same number of rows (axis_0).
A fixed random seed (42) is used for the shuffling!

Currently, two types of .h5 files are supported:

1) Files which can be read by km3pipe (e.g. files produced with OrcaSong).
2) Plain hdf5 files with a hdf5 folder depth of 1. This method is based on some legacy code.
   Be careful to not run out of memory! Needs the unshuffled .h5 file's disk space + the python overhead as memory.
   If you want to use it, please use the --legacy_mode option.
"""

import sys
import os
from argparse import ArgumentParser, RawTextHelpFormatter
import numpy as np
import h5py
import km3pipe as kp
import km3modules as km
from orcasong_contrib.data_tools.concatenate.concatenate_h5 import get_f_compression_and_chunking
import orcasong.modules as os_modules

# from memory_profiler import profile # for memory profiling, call with @profile; myfunc()

__author__ = 'Michael Moser'
__license__ = 'AGPL'
__email__ = 'michael.m.moser@fau.de'


def parse_input():
    """
    Parses the user input in order to return the most important information:

    1) list of files that should be shuffled
    2) if the unshuffled file should be deleted
    3) if the user wants to use a custom chunksize, or if the chunksize should be read from the input file.
    4) if the user wants to use a custom complib, or if the complib should be read from the input file.
    5) if the user wants to use a custom complevel, or if the complevel should be read from the input file.

    Returns
    -------
    input_files_list : list
        List that contains all filepaths of the input files that should be shuffled.
    delete : bool
        Boolean flag that specifies, if the unshuffled input files should be deleted after the shuffling.
    chunksize : None/int
        Specifies the chunksize for axis_0 in the shuffled output files.
        If None, the chunksize is read from the input files.
        Else, a custom chunksize will be used.
    complib : None/str
        Specifies the compression library that should be used for saving the shuffled output files.
        If None, the compression library is read from the input files.
        Else, a custom compression library will be used.
        Currently available: 'gzip', or 'lzf'.
    complevel : None/int
        Specifies the compression level that should be used for saving the shuffled output files.
        A compression level is only available for gzip compression, not lzf!
        If None, the compression level is read from the input files.
        Else, a custom compression level will be used.
    legacy_mode : bool
        Boolean flag that specifies, if the legacy shuffle mode should be used instead of the standard one.
        A more detailed description of this mode can be found in the summary at the top of this python file.

    """
    parser = ArgumentParser(description='E.g. < python shuffle_h5.py filepath_1 [filepath_2] [...] > \n'
                                        'Shuffles .h5 files. Requires that each dataset of the files has the same number of rows (axis_0). \n'
                                        'Outputs a new, shuffled .h5 file with the suffix < _shuffled >.',
                            formatter_class=RawTextHelpFormatter)

    parser.add_argument('files', metavar='file', type=str, nargs='+', help='a .h5 file that should be shuffled, can be more than one argument.')
    parser.add_argument('-d', '--delete', action='store_true',
                        help='deletes the original input file after the shuffled .h5 is created.')
    parser.add_argument('--chunksize', dest='chunksize', type=int,
                        help='Specify a chunksize value in order to use chunked storage for the shuffled .h5 file. \n'
                             ' Otherwise, it will be read from the input file..')
    parser.add_argument('--complib', dest='complib', type=str,
                        help='Specify a filter that should be used for compression. Either "gzip" or "lzf". \n'
                             'Otherwise, the filter will be read from the input file.')
    parser.add_argument('--complevel', dest='complevel', type=int,
                        help='Specify a compression filter strength that should be used for the compression. \n'
                             'Otherwise, the filter will be read from the input file. \n'
                             'Can range from 0 to 9. Has no effect on "lzf" compression.')
    parser.add_argument('--legacy_mode', dest='legacy_mode', action='store_true',
                        help='If you want to use the legacy mode, as described in the summary at the top of this python file.')

    parser.set_defaults(legacy_mode=False)

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    input_files_list = []
    for filepath in args.files:
        input_files_list.append(filepath)

    delete = False
    if args.delete:
        delete = True
        print('You chose delete = True')

    chunksize = None
    if args.chunksize:
        chunksize = args.chunksize
        print('You chose chunksize = ' + str(chunksize))

    complib = None
    if args.complib:
        complib = args.complib
        print('You chose complib = ' + complib)

    complevel = None
    if args.complevel:
        complevel = args.complevel
        print('You chose complevel = ' + str(complevel))

    legacy_mode = args.legacy_mode

    return input_files_list, delete, chunksize, complib, complevel, legacy_mode


def get_filepath_output(filepath_input, shuffle, event_skipper):
    """
    Get the filename of the shuffled / rebalanced output file as a str.

    """
    filepath_input_without_ext = os.path.splitext(filepath_input)[0]
    fname_adtn = ''
    if shuffle:
        fname_adtn += '_shuffled'
    if event_skipper is not None:
        fname_adtn += '_reb'
    filepath_output = filepath_input_without_ext + fname_adtn + ".h5"
    return filepath_output


def shuffle_h5(filepath_input, tool=False, seed=42, delete=False, chunksize=None,
               complib=None, complevel=None, legacy_mode=False, shuffle=True,
               event_skipper=None, filepath_output=None):
    """
    Shuffles a .h5 file where each dataset needs to have the same number of rows (axis_0).
    The shuffled data is saved to a new .h5 file with the suffix < _shuffled.h5 >.

    Can also skip certain events if a event_skipper is given.

    Parameters
    ----------
    filepath_input : str
        Filepath of the unshuffled input file.
    tool : bool
        Specifies if the function is accessed from the shuffle_h5_tool.
        In this case, the shuffled .h5 file is returned.
    seed : int
        Sets a fixed random seed for the shuffling.
    delete : bool
        Specifies if the old, unshuffled file should be deleted after extracting the data.
    chunksize : None/int
        Specifies the chunksize for axis_0 in the shuffled output files.
        If None, the chunksize is read from the input files.
        Else, a custom chunksize will be used.
    complib : None/str
        Specifies the compression library that should be used for saving the shuffled output files.
        If None, the compression library is read from the input files.
        Else, a custom compression library will be used.
        Currently available: 'gzip', or 'lzf'.
    complevel : None/int
        Specifies the compression level that should be used for saving the shuffled output files.
        A compression level is only available for gzip compression, not lzf!
        If None, the compression level is read from the input files.
        Else, a custom compression level will be used.
    legacy_mode : bool
        Boolean flag that specifies, if the legacy shuffle mode should be used instead of the standard one.
        A more detailed description of this mode can be found in the summary at the top of this python file.
    shuffle : bool
        If false, events will not be shuffled.
    event_skipper : func, optional
        Function that takes the blob as an input, and returns a bool.
        If the bool is true, the blob will be skipped.
    filepath_output : str, optional
        If given, this will be the name of the output file. Otherwise, a name
        is auto generated.

    Returns
    -------
    output_file_shuffled : h5py.File
        H5py file instance of the shuffled output file.

    """
    if event_skipper is None and not shuffle:
        raise ValueError("Either event_skipper or shuffle has to be set")

    complib_f, complevel_f, chunksize_f = get_f_compression_and_chunking(filepath_input)

    chunksize = chunksize_f if chunksize is None else chunksize
    complib = complib_f if complib is None else complib
    complevel = complevel_f if complevel is None else complevel

    if complib == 'lzf':
        complevel = None

    if filepath_output is None:
        filepath_output = get_filepath_output(filepath_input, shuffle,
                                              event_skipper)

    if not legacy_mode:
        # set random km3pipe (=numpy) seed
        print('Setting a Global Random State with the seed < 42 >.')
        km.GlobalRandomState(seed=seed)

        # km3pipe uses pytables for saving the shuffled output file, which has the name 'zlib' for the 'gzip' filter
        if complib == 'gzip':
            complib = 'zlib'

        pipe = kp.Pipeline(timeit=True)  # add timeit=True argument for profiling
        pipe.attach(km.common.StatusBar, every=200)
        pipe.attach(km.common.MemoryObserver, every=200)
        pipe.attach(kp.io.hdf5.HDF5Pump, filename=filepath_input, shuffle=shuffle, reset_index=True)
        if event_skipper is not None:
            pipe.attach(os_modules.EventSkipper, event_skipper=event_skipper)

        pipe.attach(kp.io.hdf5.HDF5Sink, filename=filepath_output, complib=complib, complevel=complevel, chunksize=chunksize, flush_frequency=1000)
        pipe.drain()

        # copy the used_files dataset to the new file
        copy_used_files(filepath_input, filepath_output)

        if delete:
            os.remove(filepath_input)

        # delete folders with '_i_' that are created by pytables in the HDF5Sink, we don't need them
        output_file_shuffled = h5py.File(filepath_output, 'r+')
        for folder_name in output_file_shuffled:
            if folder_name.startswith('_i_'):
                del output_file_shuffled[folder_name]

    else:
        input_file = h5py.File(filepath_input, 'r')
        folder_data_array_dict = {}

        for folder_name in input_file:
            folder_data_array = input_file[folder_name][()]  # get whole numpy array into memory
            folder_data_array_dict[folder_name] = folder_data_array  # workaround in order to be able to close the input file at the next step

        input_file.close()

        if delete:
            os.remove(filepath_input)

        output_file_shuffled = h5py.File(filepath_output, 'w')
        for n, dataset_key in enumerate(folder_data_array_dict):

            dataset = folder_data_array_dict[dataset_key]

            if n == 0:
                # get a particular seed for the first dataset such that the shuffling is consistent across the datasets
                r = np.random.RandomState(seed)
                state = r.get_state()
                r.shuffle(dataset)

            else:
                r.set_state(state)  # recover shuffle seed of the first dataset
                r.shuffle(dataset)

            chunks = (chunksize,) + dataset.shape[1:]
            output_file_shuffled.create_dataset(dataset_key, data=dataset, dtype=dataset.dtype, chunks=chunks,
                                                compression=complib, compression_opts=complevel)

    # close file in the case of tool=True
    if tool is False:
        output_file_shuffled.close()
    else:
        return output_file_shuffled


def copy_used_files(source_file, target_file):
    """
    Copy the "used_files" dataset from one h5 file to another, if it is present.

    """
    with h5py.File(source_file, "r") as src:
        if "used_files" in src:
            print("Copying used_files dataset to new file...")
            used_files = src["used_files"]

            with h5py.File(target_file, "a") as trg:
                trg.create_dataset("used_files", data=used_files)


def main():
    """
    Frontend for the shuffle_h5 function that can be used in a bash environment.

    Shuffles .h5 files where each dataset needs to have the same number of rows (axis_0) for a single file.
    Saves the shuffled data to a new .h5 file.
    """
    input_files_list, delete, chunksize, complib, complevel, legacy_mode = parse_input()

    for filepath_input in input_files_list:
        print('Shuffling file ' + filepath_input)
        output_file_shuffled = shuffle_h5(filepath_input, tool=True, seed=42, delete=delete, chunksize=chunksize,
                                          complib=complib, complevel=complevel, legacy_mode=legacy_mode)
        print('Finished shuffling. Output information:')
        print('---------------------------------------')
        print('The output file contains the following datasets:')
        for dataset_name in output_file_shuffled:
            print('Dataset ' + dataset_name + ' with the following shape, dtype and chunks '
                  '(first argument is the chunksize in axis_0): \n' + str(output_file_shuffled[dataset_name].shape)
                  + ' ; ' + str(output_file_shuffled[dataset_name].dtype) + ' ; '
                  + str(output_file_shuffled[dataset_name].chunks))

        output_file_shuffled.close()


if __name__ == '__main__':
    main()



import inspect
from unittest import TestCase
import orcasong.bin_edges
from orcasong.core import FileBinner


__author__ = 'Stefan Reck'


class TestEdges(TestCase):
    """
    Just call all functions in the bin_edges module and see if they work
    with the filebinner.
    """
    def test_them(self):
        funcs = [memb[1] for memb in inspect.getmembers(orcasong.bin_edges)
                 if inspect.isfunction(memb[1])]

        for func in funcs:
            fb = FileBinner(func())
            fb.get_names_and_shape()



# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from datetime import date
from pkg_resources import get_distribution

import orcasong
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "OrcaSong {}".format(orcasong.__version__)
copyright = u'{0}, Stefan Reck, Michael Moser'.format(date.today().year)
author = 'Stefan Reck, Michael Moser'

# The full version, including alpha/beta/rc tags
release = get_distribution('orcasong').version
# The short X.Y version
version = '.'.join(release.split('.')[:2])


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'autoapi.extension',
    'numpydoc',
]

autosummary_generate = True

# Document Python Code
autoapi_type = 'python'
autoapi_dirs = ['../orcasong', '../orcasong_contrib']
autoapi_options = [
    'members', 'undoc-members'
    # , 'private-members', 'special-members'
]
autoapi_ignore = [
    "*/tests/*", "*test_*.py", "*/docs/conf.py", "*/pydataskel*", "*/style/*"
]
autoapi_include_summaries = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_parsers = {
#   '.md': 'recommonmark.parser.CommonMarkParser',}
# source_suffix = ['.rst', '.md']
source_suffix = ['.rst']

# The master toctree document.
master_doc = 'index'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_theme_options = {
    'logo_only': True,
}
html_logo = "_static/orcasong_wide_transparent_white.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = "OrcaSong {}".format(orcasong.__version__)


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}


# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'OrcaSongdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'OrcaSong.tex', 'OrcaSong Documentation',
     'Stefan Reck, Michael Moser', 'manual'),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'orcasong', 'OrcaSong Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'OrcaSong', 'OrcaSong Documentation',
     author, 'OrcaSong', 'One line description of project.',
     'Miscellaneous'),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


def setup(app):
    app.add_stylesheet('_static/style.css')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utility script that makes .list files for the concatenate_h5.py tool.

Usage:
    make_data_split.py CONFIG
    make_data_split.py (-h | --help)

Arguments:
    CONFIG  A .toml file which contains the configuration options.

Options:
    -h --help  Show this screen.

"""

__author__ = 'Michael Moser'

import os
import toml
import docopt
import natsort as ns
import h5py
import random

def parse_input():
    """
    Parses the config of the .toml file, specified by the user.

    Returns
    -------
    cfg : dict
        Dict that contains all configuration options from the input .toml file.

    """

    args = docopt.docopt(__doc__)
    config_file = args['CONFIG']

    cfg = toml.load(config_file)
    cfg['toml_filename'] = config_file

    if 'chunksize' not in cfg: cfg['chunksize'] = None
    if 'complib' not in cfg: cfg['complib'] = None
    if 'complevel' not in cfg: cfg['complevel'] = None
    return cfg


def get_all_ip_group_keys(cfg):
    """
    Gets the keys of all input groups in the config dict.

    The input groups are defined as the dict elements, where the values have the type of a dict.

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.

    Returns
    -------
    ip_group_keys : list
        List of the input_group keys.

    """
    ip_group_keys = []
    for key in cfg:
        if type(cfg[key]) == dict:
            ip_group_keys.append(key)

    return ip_group_keys


def get_h5_filepaths(dirpath):
    """
    Returns the filepaths of all .h5 files that are located in a specific directory.

    Parameters
    ----------
    dirpath: str
        Path of the directory where the .h5 files are located.

    Returns
    -------
    filepaths : list
        List with the full filepaths of all .h5 files in the dirpath folder.

    """
    filepaths = []
    for f in os.listdir(dirpath):
        if f.endswith('.h5'):
            filepaths.append(dirpath + '/' + f)

    #randomize order
    random.Random(42).shuffle(filepaths)
    
    return filepaths


def get_number_of_evts_and_run_ids(list_of_files, dataset_key='y', run_id_col_name='run_id'):
    """
    Gets the number of events and the run_ids for all hdf5 files in the list_of_files.

    The number of events is calculated based on the dataset, which is specified with the dataset_key parameter.

    Parameters
    ----------
    list_of_files : list
        List which contains filepaths to h5 files.
    dataset_key : str
        String which specifies, which dataset in a h5 file should be used for calculating the number of events.
    run_id_col_name : str
        String, which specifies the column name of the 'run_id' column.

    Returns
    -------
    total_number_of_evts : int
        The cumulative (total) number of events.
    mean_number_of_evts_per_file : float
        The mean number of evts per file.
    run_ids : list
        List containing the run_ids of the files in the list_of_files.

    """

    total_number_of_evts = 0
    run_ids = []

    for i, fpath in enumerate(list_of_files):
        f = h5py.File(fpath, 'r')

        dset = f[dataset_key]
        n_evts = dset.shape[0]
        total_number_of_evts += n_evts

        run_id = f[dataset_key][0][run_id_col_name]
        run_ids.append(run_id)

        f.close()

    mean_number_of_evts_per_file = total_number_of_evts / len(list_of_files)

    return total_number_of_evts, mean_number_of_evts_per_file, run_ids


def split(a, n):
    """
    Splits a list into n equal sized (if possible! if not, approximately) chunks.

    Parameters
    ----------
    a : list
        A list that should be split.
    n : int
        Number of times the input list should be split.

    Returns
    -------
    a_split : list
        The input list a, which has been split into n chunks.

    """
    # from https://stackoverflow.com/questions/2130016/splitting-a-list-into-n-parts-of-approximately-equal-length
    k, m = divmod(len(a), n)
    a_split = list((a[i * k + min(i, m):(i + 1) * k + min(i + 1, m)] for i in range(n)))
    return a_split


def print_input_statistics(cfg, ip_group_keys):
    """
    Prints some useful information for each input_group.

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.
    ip_group_keys : list
        List of the input_group keys.

    """

    print('----------------------------------------------------------------------')
    print('Printing input statistics for your ' + cfg['toml_filename'] + ' input:')
    print('----------------------------------------------------------------------')

    print('Your input .toml file has the following data input groups: ' + str(ip_group_keys))
    print('Total number of events: ' + str(cfg['n_evts_total']))

    for key in ip_group_keys:
        print('--------------------------------------------------------------------')
        print('Info for group ' + key + ':')
        print('Directory: ' + cfg[key]['dir'])
        print('Total number of files: ' + str(cfg[key]['n_files']))
        print('Total number of events: ' + str(cfg[key]['n_evts']))
        print('Mean number of events per file: ' + str(round(cfg[key]['n_evts_per_file_mean'], 3)))
        print('--------------------------------------------------------------------')


def add_fpaths_for_data_split_to_cfg(cfg, key):
    """
    Adds all the filepaths for the output files into a list, and puts them into the cfg['output_dsplit'][key] location
    for all dsplits (train, validate, rest).

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.
    key : str
        The key of an input_group.

    """

    fpath_lists = {'train': [], 'validate': [], 'rest': []}
    for i, fpath in enumerate(cfg[key]['fpaths']):

        run_id = cfg[key]['run_ids'][i]

        for dsplit in ['train', 'validate', 'rest']:
            if 'run_ids_' + dsplit in cfg[key]:
                if cfg[key]['run_ids_' + dsplit][0] <= run_id <= cfg[key]['run_ids_' + dsplit][1]:
                    fpath_lists[dsplit].append(fpath)

    for dsplit in ['train', 'validate', 'rest']:
        if len(fpath_lists[dsplit]) == 0:
            continue

        n_files_dsplit = cfg['n_files_' + dsplit]
        fpath_lists[dsplit] = split(fpath_lists[dsplit], n_files_dsplit)
        if 'output_' + dsplit not in cfg:
            cfg['output_' + dsplit] = dict()
        cfg['output_' + dsplit][key] = fpath_lists[dsplit]


def make_dsplit_list_files(cfg):
    """
    Writes .list files of the datasplits to the disk, with the information in the cfg['output_dsplit'] dict.

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.

    """
    # check if //conc_list_files folder exists, if not create it.
    if not os.path.exists(cfg['output_file_folder'] + '/conc_list_files'):
        os.makedirs(cfg['output_file_folder'] + '/conc_list_files')

    for dsplit in ['train', 'validate', 'rest']:

        if 'output_' + dsplit not in cfg:
            continue

        first_key = list(cfg['output_' + dsplit].keys())[0]
        n_output_files = len(cfg['output_' + dsplit][first_key])

        for i in range(n_output_files):
            fpath_output = cfg['output_file_folder'] + '/conc_list_files/' + cfg['output_file_name'] + '_' + dsplit + '_' + str(i) + '.list'

            # for later usage
            if 'output_lists' not in cfg:
                cfg['output_lists'] = list()
            cfg['output_lists'].append(fpath_output)

            with open(fpath_output, 'w') as f_out:
                for group_key in cfg['output_' + dsplit]:
                    for fpath in cfg['output_' + dsplit][group_key][i]:
                        f_out.write(fpath + '\n')


def make_concatenate_and_shuffle_list_files(cfg):
    """
    Function that writes qsub .sh files which concatenates all files inside the .list files.

    Parameters
    ----------
    cfg : dict
        Dict that contains all configuration options and additional information.

    """
    # TODO include options for multicore

    dirpath = cfg['output_file_folder']

    if not os.path.exists(dirpath + '/logs'):  # check if /logs folder exists, if not create it.
        os.makedirs(dirpath + '/logs')
    if not os.path.exists(dirpath + '/job_scripts'):  # check if /job_scripts folder exists, if not create it.
        os.makedirs(dirpath + '/job_scripts')
    if not os.path.exists(dirpath + '/data_split'):  # check if /data_split folder exists, if not create it.
        os.makedirs(dirpath + '/data_split')

    chunksize = '' if cfg['chunksize'] is None else ' --chunksize ' + str(cfg['chunksize'])
    complib = '' if cfg['complib'] is None else ' --complib ' + str(cfg['complib'])
    complevel = '' if cfg['complevel'] is None else ' --complevel ' + str(cfg['complevel'])

    # make qsub .sh file for concatenating
    for listfile_fpath in cfg['output_lists']:
        listfile_fname = os.path.basename(listfile_fpath)
        listfile_fname_wout_ext = os.path.splitext(listfile_fname)[0]
        conc_outputfile_fpath = cfg['output_file_folder'] + '/data_split/' + listfile_fname_wout_ext + '.h5'

        fpath_bash_script = dirpath + '/job_scripts/concatenate_h5_' + listfile_fname_wout_ext + '.sh'

        with open(fpath_bash_script, 'w') as f:
            f.write('#!/usr/bin/env bash\n')
            f.write('#\n')
            f.write('#PBS -o ' + cfg['output_file_folder'] + '/logs/concatenate_h5_' + listfile_fname_wout_ext + '.out'
                    ' -e ' + cfg['output_file_folder'] + '/logs/concatenate_h5_' + listfile_fname_wout_ext + '.err\n')
            f.write('\n')
            f.write('CodeFolder="' + cfg['data_tools_folder'] + '"\n')
            f.write('cd ${CodeFolder}\n')
            f.write('source activate ' + cfg['venv_path'] + '\n')
            f.write('\n')
            f.write('# Concatenate the files in the list\n')

            f.write('time python concatenate/concatenate_h5.py'
                    + chunksize + complib + complevel
                    + ' -l ' + listfile_fpath + ' ' + conc_outputfile_fpath)

        if cfg['submit_jobs'] is True:
            os.system('qsub -l nodes=1:ppn=4,walltime=23:59:00 ' + fpath_bash_script)

    # make qsub .sh file for shuffling
    delete_flag_shuffle_tool = ' --delete' if cfg['shuffle_delete'] is True else ''
    for listfile_fpath in cfg['output_lists']:
        listfile_fname = os.path.basename(listfile_fpath)
        listfile_fname_wout_ext = os.path.splitext(listfile_fname)[0]

        # This is the input for the shuffle tool!
        conc_outputfile_fpath = cfg['output_file_folder'] + '/data_split/' + listfile_fname_wout_ext + '.h5'

        fpath_bash_script = dirpath + '/job_scripts/shuffle_h5_' + listfile_fname_wout_ext + '.sh'

        with open(fpath_bash_script, 'w') as f:
            f.write('#!/usr/bin/env bash\n')
            f.write('#\n')
            f.write('#PBS -o ' + cfg['output_file_folder'] + '/logs/shuffle_h5_' + listfile_fname_wout_ext + '.out'
                    ' -e ' + cfg['output_file_folder'] + '/logs/shuffle_h5_' + listfile_fname_wout_ext + '.err\n')
            f.write('\n')
            f.write('CodeFolder="' + cfg['data_tools_folder'] + '"\n')
            f.write('cd ${CodeFolder}\n')
            f.write('source activate ' + cfg['venv_path'] + '\n')
            f.write('\n')
            f.write('# Shuffle the h5 file \n')

            f.write('time python shuffle/shuffle_h5.py'
                    + delete_flag_shuffle_tool
                    + chunksize + complib + complevel
                    + ' ' + conc_outputfile_fpath)


def main():
    """
    Main function to make the data split.
    """

    cfg = parse_input()

    ip_group_keys = get_all_ip_group_keys(cfg)

    n_evts_total = 0
    for key in ip_group_keys:
        print('Collecting information from input group ' + key)
        cfg[key]['fpaths'] = get_h5_filepaths(cfg[key]['dir'])
        cfg[key]['n_files'] = len(cfg[key]['fpaths'])
        cfg[key]['n_evts'], cfg[key]['n_evts_per_file_mean'], cfg[key]['run_ids'] = get_number_of_evts_and_run_ids(cfg[key]['fpaths'], dataset_key='y')

        n_evts_total += cfg[key]['n_evts']

    cfg['n_evts_total'] = n_evts_total
    print_input_statistics(cfg, ip_group_keys)

    if cfg['print_only'] is True:
        from sys import exit
        exit()

    for key in ip_group_keys:
        add_fpaths_for_data_split_to_cfg(cfg, key)

    make_dsplit_list_files(cfg)

    if cfg['make_qsub_bash_files'] is True:
        make_concatenate_and_shuffle_list_files(cfg)


if __name__ == '__main__':
    main()



"""
Functions that extract info from a blob for the mc_info / y datafield
in the h5 files.

These are made for the specific given runs. They might not be
applicable to other data, and could cause errors or produce unexpected
results when used on data other then the specified.

"""

import warnings
import numpy as np
from km3pipe.io.hdf5 import HDF5Header
from h5py import File

__author__ = "Daniel Guderian"


def get_std_reco(blob):

    """
    Function to extract std reco info. The implemented strategy is the following:
    First, look for whether a rec stag has been reached and only then extract the reconstructed
    paramater from it. If not, set it to a dummy value (for now 0). This means that for an analysis the events with
    exactly zero have to be filtered out!

    The 'best track' is the first (highest lik) while a certain rec stage has to be reached. This might
    have to be adjusted for other recos than JMuonGandalf chain.


    Members of the Tracks:
    dtype([('E', '<f8'), ('JCOPY_Z_M', '<f4'), ('JENERGY_CHI2', '<f4'), ('JENERGY_ENERGY', '<f4'),
     ('JENERGY_MUON_RANGE_METRES', '<f4'), ('JENERGY_NDF', '<f4'), ('JENERGY_NOISE_LIKELIHOOD', '<f4'),
     ('JENERGY_NUMBER_OF_HITS', '<f4'), ('JGANDALF_BETA0_RAD', '<f4'), ('JGANDALF_BETA1_RAD', '<f4'),
     ('JGANDALF_CHI2', '<f4'), ('JGANDALF_LAMBDA', '<f4'), ('JGANDALF_NUMBER_OF_HITS', '<f4'),
     ('JGANDALF_NUMBER_OF_ITERATIONS', '<f4'), ('JSHOWERFIT_ENERGY', '<f4'), ('JSTART_LENGTH_METRES', '<f4'),
     ('JSTART_NPE_MIP', '<f4'), ('JSTART_NPE_MIP_TOTAL', '<f4'), ('JVETO_NPE', '<f4'), ('JVETO_NUMBER_OF_HITS', '<f4'),
     ('dir_x', '<f8'), ('dir_y', '<f8'), ('dir_z', '<f8'), ('id', '<i4'), ('idx', '<i8'), ('length', '<f8'),
     ('likelihood', '<f8'), ('pos_x', '<f8'), ('pos_y', '<f8'), ('pos_z', '<f8'), ('rec_type', '<i4'),
     ('t', '<f8'), ('group_id', '<i8')])


    members of rec stages:
    .idx (corresponding to the track id),
    .rec_stage (rec stage identifier, for JMuonGandalf for example: 1=prefit, 2=simplex, 3=gandalf,
                    4=engery, 5=start),
    .group_id (event id in file)


    Parameters
    ----------
    blob : blob containing the reco info

    Returns
    -------
    std_reco_info : dict
            Dict with the most common std reco params. Can be expanded.

    """

    # use this later to identify not reconstructed events
    dummy_value = 0

    # if there was no std reco at all, this will not exist
    # these are events that stopped at/before prefit
    try:
        rec_stages = blob["RecStages"]
        # get first track only
        rec_stages_best_track = rec_stages.rec_stage[rec_stages.idx == 0]

        # often enough: best track is the first
        best_track = blob["Tracks"][0]

    except KeyError:
        rec_stages_best_track = []
        print(
            "An event didnt have any reco. Setting everything to"
            + str(dummy_value)
            + "."
        )

    # take the direction only if JGanalf was executed
    if 3 in rec_stages_best_track:

        std_dir_x = best_track["dir_x"]
        std_dir_y = best_track["dir_y"]
        std_dir_z = best_track["dir_z"]

        std_beta0 = best_track["JGANDALF_BETA0_RAD"]
        std_lik = best_track["likelihood"]
        std_n_hits_gandalf = best_track["JGANDALF_NUMBER_OF_HITS"]

    else:

        std_dir_x = dummy_value
        std_dir_y = dummy_value
        std_dir_z = dummy_value

        std_beta0 = dummy_value
        std_lik = dummy_value
        std_n_hits_gandalf = dummy_value

    # energy fit from JEnergy
    if 4 in rec_stages_best_track:

        std_energy = best_track["E"]
        lik_energy = best_track["JENERGY_CHI2"]

    else:
        std_energy = dummy_value
        lik_energy = dummy_value

    # vertex and length from JStart
    if 5 in rec_stages_best_track:

        std_pos_x = best_track["pos_x"]
        std_pos_y = best_track["pos_y"]
        std_pos_z = best_track["pos_z"]

        std_length = best_track["JSTART_LENGTH_METRES"]

    else:

        std_pos_x = dummy_value
        std_pos_y = dummy_value
        std_pos_z = dummy_value

        std_length = dummy_value

    std_reco_info = {
        "std_dir_x": std_dir_x,
        "std_dir_y": std_dir_y,
        "std_dir_z": std_dir_z,
        "std_beta0": std_beta0,
        "std_lik": std_lik,
        "std_n_hits_gandalf": std_n_hits_gandalf,
        "std_pos_x": std_pos_x,
        "std_pos_y": std_pos_y,
        "std_pos_z": std_pos_z,
        "std_energy": std_energy,
        "std_lik_energy": lik_energy,
        "std_length": std_length,
    }

    return std_reco_info


def get_real_data_info_extr(input_file):

    """
    Wrapper function that includes the actual mc_info_extr
    for real data. There are no n_gen like in the neutrino case.

    Parameters
    ----------
    input_file : km3net data file
            Can be online or offline format.

    Returns
    -------
    mc_info_extr : function
            The actual mc_info_extr function that holds the extractions.

    """

    # check if std reco is present
    f = File(input_file, "r")
    has_std_reco = "reco" in f.keys()

    def mc_info_extr(blob):

        """
        Processes a blob and creates the y with mc_info and, if existing, std reco.

        For this real data case it is only general event info, like the id.

        Parameters
        ----------
        blob : dict
                The blob from the pipeline.

        Returns
        -------
        track : dict
                Containing all the specified info the y should have.

        """

        event_info = blob["EventInfo"][0]

        # add n_hits info for the cut
        n_hits = len(blob["Hits"])

        track = {
            "event_id": event_info.event_id,
            "run_id": event_info.run_id,
            "trigger_mask": event_info.trigger_mask,
            "n_hits": n_hits,
        }

        # get all the std reco info
        if has_std_reco:

            std_reco_info = get_std_reco(blob)

            track.update(std_reco_info)

        return track

    return mc_info_extr


def get_random_noise_mc_info_extr(input_file):

    """
    Wrapper function that includes the actual mc_info_extr
    for random noise simulations. There are no n_gen like in the neutrino case.

    Parameters
    ----------
    input_file : km3net data file
            Can be online or offline format.

    Returns
    -------
    mc_info_extr : function
            The actual mc_info_extr function that holds the extractions.

    """

    # check if std reco is present
    f = File(input_file, "r")
    has_std_reco = "reco" in f.keys()

    def mc_info_extr(blob):

        """
        Processes a blob and creates the y with mc_info and, if existing, std reco.

        For this random noise case it is only general event info, like the id.

        Parameters
        ----------
        blob : dict
                The blob from the pipeline.

        Returns
        -------
        track : dict
                Containing all the specified info the y should have.

        """
        event_info = blob["EventInfo"]

        track = {
            "event_id": event_info.event_id[0],
            "run_id": event_info.run_id[0],
            "particle_type": 0,
        }

        # get all the std reco info
        if has_std_reco:

            std_reco_info = get_std_reco(blob)

            track.update(std_reco_info)

        return track

    return mc_info_extr


def get_neutrino_mc_info_extr(input_file):

    """
    Wrapper function that includes the actual mc_info_extr
    for neutrino simulations. The n_gen parameter, needed for neutrino weighting
    is extracted from the header of the file.

    Parameters
    ----------
    input_file : km3net data file
            Can be online or offline format.

    Returns
    -------
    mc_info_extr : function
            The actual mc_info_extr function that holds the extractions.

    """

    # check if std reco is present
    f = File(input_file, "r")
    has_std_reco = "reco" in f.keys()

    # get the n_gen
    header = HDF5Header.from_hdf5(input_file)
    n_gen = header.genvol.numberOfEvents

    def mc_info_extr(blob):

        """
        Processes a blob and creates the y with mc_info and, if existing, std reco.

        For this neutrino case it is the full mc info for the primary neutrino; there are the several "McTracks":
        check the simulation which index "p" the neutrino has.

        Parameters
        ----------
        blob : dict
                The blob from the pipeline.

        Returns
        -------
        track : dict
                Containing all the specified info the y should have.

        """

        # get general info about the event
        event_id = blob["EventInfo"].event_id[0]
        run_id = blob["EventInfo"].run_id[0]
        # weights for neutrino analysis
        weight_w1 = blob["EventInfo"].weight_w1[0]
        weight_w2 = blob["EventInfo"].weight_w2[0]
        weight_w3 = blob["EventInfo"].weight_w3[0]

        # first, look for the particular neutrino index of the production
        p = 0  # for ORCA4 (and probably subsequent productions)

        mc_track = blob["McTracks"][p]

        # some track mc truth info
        particle_type = mc_track.type
        energy = mc_track.energy
        is_cc = mc_track.cc
        bjorkeny = mc_track.by
        dir_x, dir_y, dir_z = mc_track.dir_x, mc_track.dir_y, mc_track.dir_z
        time_interaction = (
            mc_track.time
        )  # actually always 0 for primary neutrino, measured in MC time
        vertex_pos_x, vertex_pos_y, vertex_pos_z = (
            mc_track.pos_x,
            mc_track.pos_y,
            mc_track.pos_z,
        )

        # add also the nhits info
        n_hits = len(blob["Hits"])

        track = {
            "event_id": event_id,
            "particle_type": particle_type,
            "energy": energy,
            "is_cc": is_cc,
            "bjorkeny": bjorkeny,
            "dir_x": dir_x,
            "dir_y": dir_y,
            "dir_z": dir_z,
            "time_interaction": time_interaction,
            "run_id": run_id,
            "vertex_pos_x": vertex_pos_x,
            "vertex_pos_y": vertex_pos_y,
            "vertex_pos_z": vertex_pos_z,
            "n_hits": n_hits,
            "weight_w1": weight_w1,
            "weight_w2": weight_w2,
            "weight_w3": weight_w3,
            "n_gen": n_gen,
        }
        # get all the std reco info
        if has_std_reco:

            std_reco_info = get_std_reco(blob)

            track.update(std_reco_info)

        return track

    return mc_info_extr


def get_muon_mc_info_extr(input_file):

    """
    Wrapper function that includes the actual mc_info_extr
    for atm. muon simulations. There are no n_gen like in the neutrino case.

    Parameters
    ----------
    input_file : km3net data file
            Can be online or offline format.

    Returns
    -------
    mc_info_extr : function
            The actual mc_info_extr function that holds the extractions.

    """

    # check if std reco is present
    f = File(input_file, "r")
    has_std_reco = "reco" in f.keys()

    # no n_gen here, but needed for concatenation
    n_gen = 1

    def mc_info_extr(blob):

        """
        Processes a blob and creates the y with mc_info and, if existing, std reco.

        For this atm. muon case it is the full mc info for the primary; there are the several "McTracks":
        check the simulation to understand what "p" you want. Muons come in bundles that have the same direction.
        For energy: sum of all muons in a bundle,
        for vertex: weighted (energy) mean of the individual vertices .

        Parameters
        ----------
        blob : dict
                The blob from the pipeline.

        Returns
        -------
        track : dict
                Containing all the specified info the y should have.

        """

        event_id = blob["EventInfo"].event_id[0]
        run_id = blob["EventInfo"].run_id[0]

        p = 0  # for ORCA4 (and probably subsequent productions)

        mc_track = blob["McTracks"][p]

        particle_type = (
            mc_track.type
        )  # assumed that this is the same for all muons in a bundle
        is_cc = mc_track.cc  # always 0 actually
        bjorkeny = mc_track.by
        time_interaction = mc_track.time  # same for all muons in a bundle

        # sum up the energy of all muons
        energy = np.sum(blob["McTracks"].energy)

        # all muons in a bundle are parallel, so just take dir of first muon
        dir_x, dir_y, dir_z = mc_track.dir_x, mc_track.dir_y, mc_track.dir_z

        # vertex is the weighted (energy) mean of the individual vertices
        vertex_pos_x = np.average(
            blob["McTracks"][p:].pos_x, weights=blob["McTracks"][p:].energy
        )
        vertex_pos_y = np.average(
            blob["McTracks"][p:].pos_y, weights=blob["McTracks"][p:].energy
        )
        vertex_pos_z = np.average(
            blob["McTracks"][p:].pos_z, weights=blob["McTracks"][p:].energy
        )

        # add also the nhits info
        n_hits = len(blob["Hits"])

        # this is only relevant for neutrinos, though dummy info is needed for the concatenation
        weight_w1 = 10
        weight_w2 = 10
        weight_w3 = 10

        track = {
            "event_id": event_id,
            "particle_type": particle_type,
            "energy": energy,
            "is_cc": is_cc,
            "bjorkeny": bjorkeny,
            "dir_x": dir_x,
            "dir_y": dir_y,
            "dir_z": dir_z,
            "time_interaction": time_interaction,
            "run_id": run_id,
            "vertex_pos_x": vertex_pos_x,
            "vertex_pos_y": vertex_pos_y,
            "vertex_pos_z": vertex_pos_z,
            "n_hits": n_hits,
            "weight_w1": weight_w1,
            "weight_w2": weight_w2,
            "weight_w3": weight_w3,
            "n_gen": n_gen,
        }

        # get all the std reco info
        if has_std_reco:

            std_reco_info = get_std_reco(blob)

            track.update(std_reco_info)

        return track

    return mc_info_extr



"""
Scripts for postprocessing h5 files, e.g. shuffling.
"""
import os
import argparse
import h5py
import km3pipe as kp
import km3modules as km
from orcasong.modules import EventSkipper
from orcasong.tools.concatenate import get_compopts, copy_attrs


def postproc_file(
        input_file,
        output_file=None,
        shuffle=True,
        event_skipper=None,
        delete=False,
        seed=42,
        statusbar_every=1000):
    """
    Postprocess a file using km3pipe after it has been preprocessed in OrcaSong.

    Parameters
    ----------
    input_file : str
        Path of the file that will be processed.
    output_file : str, optional
        If given, this will be the name of the output file.
        Otherwise, a name is auto generated.
    shuffle : bool
        Shuffle order of events.
    event_skipper : func, optional
        Function that takes the blob as an input, and returns a bool.
        If the bool is true, the event will be skipped.
    delete : bool
        Specifies if the input file should be deleted after processing.
    seed : int
        Sets a fixed random seed for the shuffling.
    statusbar_every : int or None
        After how many line a km3pipe status should be printed.

    Returns
    -------
    output_file : str
        Path to the output file.

    """
    if output_file is None:
        output_file = get_filepath_output(
            input_file, shuffle=shuffle, event_skipper=event_skipper)
    if os.path.exists(output_file):
        raise FileExistsError(output_file)

    print(f'Setting a Global Random State with the seed < {seed} >.')
    km.GlobalRandomState(seed=seed)

    comptopts = get_compopts(input_file)
    # km3pipe uses pytables for saving the shuffled output file,
    # which has the name 'zlib' for the 'gzip' filter
    if comptopts["complib"] == 'gzip':
        comptopts["complib"] = 'zlib'

    pipe = kp.Pipeline()
    if statusbar_every is not None:
        pipe.attach(km.common.StatusBar, every=statusbar_every)
        pipe.attach(km.common.MemoryObserver, every=statusbar_every)
    pipe.attach(
        kp.io.hdf5.HDF5Pump,
        filename=input_file,
        shuffle=shuffle,
        reset_index=True,
    )
    if event_skipper is not None:
        pipe.attach(EventSkipper, event_skipper=event_skipper)
    pipe.attach(
        kp.io.hdf5.HDF5Sink,
        filename=output_file,
        complib=comptopts["complib"],
        complevel=comptopts["complevel"],
        chunksize=comptopts["chunksize"],
        flush_frequency=1000,
    )
    pipe.drain()

    copy_used_files(input_file, output_file)
    copy_attrs(input_file, output_file)
    if delete:
        print("Deleting original file")
        os.remove(input_file)

    print("Done!")
    return output_file


def copy_used_files(source_file, target_file):
    """ Copy the "used_files" dataset from one h5 file to another, if it is present.
    """
    with h5py.File(source_file, "r") as src:
        if "used_files" in src:
            print("Copying used_files dataset")
            with h5py.File(target_file, "a") as trg:
                trg.create_dataset("used_files", data=src["used_files"])


def get_filepath_output(input_file, shuffle=True, event_skipper=None):
    """ Get the filename of the shuffled / rebalanced output file as a str.
    """
    fname_adtn = ''
    if shuffle:
        fname_adtn += '_shuffled'
    if event_skipper is not None:
        fname_adtn += '_reb'
    return f"{os.path.splitext(input_file)[0]}{fname_adtn}.h5"


def h5shuffle():
    parser = argparse.ArgumentParser(description='Shuffle an h5 file using km3pipe.')
    parser.add_argument('input_file', type=str, help='File to shuffle.')
    parser.add_argument('--output_file', type=str,
                        help='Name of output file. Default: Auto generate name.')
    parser.add_argument('--delete', action="store_true",
                        help='Delete original file afterwards.')

    postproc_file(**vars(parser.parse_args()), shuffle=True, event_skipper=None)



"""
Binnings used for some existing detector configurations.

These are made for the specific given runs. They might not be
applicable to other data, and could cause errors or produce unexpected
results when used on data other then the specified.
"""

import numpy as np


def get_4line_bin_edges_list(time_resolution="low"):
    """
    Designed for four line detector.

    XYZ/PM : 1 dom per bin
    Time:
        low:
            100 bins, 6 ns/bin
            mchits cut off (left/right): 1.34%, 4.20%
        high:
            300 bins, 3.33 ns/bin
            mchits cut off (left/right): 0.6%, 0.73%

    """
    time_resolutions = {
        "low": np.linspace(-50, 550, 100 + 1),
        "high": np.linspace(-100, 900, 300 + 1),
    }

    return [
        ["pos_x", np.array([-2 - 30, -2, -2 + 30])],
        ["pos_y", np.array([3.7 - 30, 3.7, 3.7 + 30])],
        ["pos_z", np.linspace(24, 198, 19)],
        ["time", time_resolutions[time_resolution]],
        ["channel_id", np.linspace(-0.5, 30.5, 31 + 1)],
    ]


def get_edges_2017_ztc():
    """
    Designed for the 2017 runs with the one line detector.

    Will produce (18, 100, 31) 3d data, with dimensions ztc.

    Z binning: 9.45 meters each
    Time binning: 6 ns each
    Channel id binning: 1 DOM per bin

    """
    bin_edges_list = [
        ["pos_z", np.linspace(26, 198, 18 + 1)],
        ["time", np.linspace(-50, 550, 100 + 1)],
        ["channel_id", np.linspace(-0.5, 30.5, 31 + 1)],
    ]
    return bin_edges_list



#!/usr/bin/env python
# coding=utf-8
# Filename: make_nn_images.py
"""
Main OrcaSong code which takes raw simulated .h5 files and the corresponding .detx detector file as input in
order to generate 2D/3D/4D histograms ('images') that can be used for CNNs.

First argument: KM3NeT hdf5 simfile at JTE level.
Second argument: a .detx file that is associated with the hdf5 file.

The input file can be calibrated or not (e.g. contains pos_xyz of the hits).

Usage:
    make_nn_images.py SIMFILE DETXFILE CONFIGFILE
    make_nn_images.py (-h | --help)

Arguments:
    FILENAME    A KM3NeT hdf5 simfile at JTE level.

    DETXFILE    A .detx geometry file that is associated with the hdf5 file.

    CONFIGFILE  A .toml configuration file that contains all configuration options
                of this script. A default config can be found in the OrcaSong repo:
                orcasong/default_config.toml

Options:
    -h --help  Show this screen.

"""

__author__ = 'Michael Moser'
__license__ = 'AGPL'
__version__ = '1.0'
__email__ = 'michael.m.moser@fau.de'
__status__ = 'Prototype'

import warnings
import os
import sys
#from memory_profiler import profile # for memory profiling, call with @profile; myfunc()
#import line_profiler # call with kernprof -l -v file.py args
import km3pipe as kp
import km3modules as km
import matplotlib as mpl
from docopt import docopt
mpl.use('Agg')
from matplotlib.backends.backend_pdf import PdfPages

from legacy.file_to_hits import EventDataExtractor
from legacy.hits_to_histograms import HistogramMaker
from legacy.io import load_config, check_user_input, make_output_dirs
from legacy.geo_binning import calculate_bin_edges
from legacy.utils import get_file_particle_type, EventSkipper


# TODO deprecated
warnings.warn("The original Orcasong is deprecated, and is no longer supported. "
              "Consider switching to the new orcasong.")


def parse_input():
    """
    Parses and returns all necessary input options for the make_nn_images function.

    Returns
    -------
    fname : str
        Full filepath to the input .h5 file.
    detx_filepath : str
        Full filepath to the .detx geometry file that belongs to the fname.
    config_filepath : str
        Full filepath to a config file. An example can be found in orcasong/default_config.toml

    """
    args = docopt(__doc__)

    fname = args['SIMFILE']
    detx_filepath = args['DETXFILE']
    config_filepath = args['CONFIGFILE']

    return fname, detx_filepath, config_filepath


def make_nn_images(fname, detx_filepath, config):
    """
    Main code with config parameters. Reads raw .hdf5 files and creates 2D/3D histogram projections that can be used
    for a CNN.

    Parameters
    ----------
    fname : str
        Filename (full path!) of the input file.
    detx_filepath : str
        String with the full filepath to the corresponding .detx file of the input file.
        Used for the binning and for the hits calibration if the input file is not calibrated yet
        (e.g. hits do not contain pos_x/y/z, time, ...).
    config : dict
        Dictionary that contains all configuration options of the make_nn_images function.
        An explanation of the config parameters can be found in orcasong/default_config.toml.

    """
    # Load all parameters from the config # TODO put everything in a config class, this is horrible
    output_dirpath = config['output_dirpath']
    chunksize, complib, complevel = config['chunksize'], config['complib'], config['complevel']
    flush_freq = config['flush_freq']
    n_bins = tuple(config['n_bins'])
    timecut = (config['timecut_mode'], config['timecut_timespan'])
    do_mc_hits = config['do_mc_hits']
    det_geo = config['det_geo']
    do2d = config['do2d']
    do2d_plots = (config['do2d_plots'], config['do2d_plots_n'])
    do3d = config['do3d']
    do4d = (config['do4d'], config['do4d_mode'])
    prod_ident = config['prod_ident'] if config['prod_ident'] != 'None' else None
    data_cuts = dict()
    data_cuts['triggered'] = config['data_cut_triggered']
    data_cuts['energy_lower_limit'] = config['data_cut_e_low'] if config['data_cut_e_low'] != 'None' else None
    data_cuts['energy_upper_limit'] = config['data_cut_e_high'] if config['data_cut_e_high'] != 'None' else None
    data_cuts['throw_away_prob'] = config['data_cut_throw_away'] if config['data_cut_throw_away'] != 'None' else None
    data_cuts['custom_skip_function'] = config['data_cut_custom_func'] if config['data_cut_custom_func'] != 'None' else None

    make_output_dirs(output_dirpath, do2d, do3d, do4d)

    filename = os.path.basename(os.path.splitext(fname)[0])
    filename_output = filename.replace('.','_')

    # set random km3pipe (=numpy) seed
    print('Setting a Global Random State with the seed < 42 >.')
    km.GlobalRandomState(seed=42)

    geo, x_bin_edges, y_bin_edges, z_bin_edges = calculate_bin_edges(n_bins, det_geo, detx_filepath, do4d)
    pdf_2d_plots = PdfPages(output_dirpath + '/orcasong_output/4dTo2d/' + filename_output + '_plots.pdf') if do2d_plots[0] is True else None

    file_particle_type = get_file_particle_type(fname)

    print('Generating histograms from the hits for files based on ' + fname)

    # Initialize OrcaSong Event Pipeline

    pipe = kp.Pipeline() # add timeit=True argument for profiling
    pipe.attach(km.common.StatusBar, every=200)
    pipe.attach(km.common.MemoryObserver, every=400)
    pipe.attach(kp.io.hdf5.HDF5Pump, filename=fname)
    pipe.attach(km.common.Keep, keys=['EventInfo', 'Header', 'RawHeader', 'McTracks', 'Hits', 'McHits'])
    pipe.attach(EventDataExtractor,
                file_particle_type=file_particle_type, geo=geo, do_mc_hits=do_mc_hits,
                data_cuts=data_cuts, do4d=do4d, prod_ident=prod_ident)
    pipe.attach(km.common.Keep, keys=['event_hits', 'event_track'])
    pipe.attach(EventSkipper, data_cuts=data_cuts)
    pipe.attach(HistogramMaker,
                x_bin_edges=x_bin_edges, y_bin_edges=y_bin_edges, z_bin_edges=z_bin_edges,
                n_bins=n_bins, timecut=timecut, do2d=do2d, do2d_plots=do2d_plots, pdf_2d_plots=pdf_2d_plots,
                do3d=do3d, do4d=do4d)
    pipe.attach(km.common.Delete, keys=['event_hits'])

    if do2d:
        for proj in ['xy', 'xz', 'yz', 'xt', 'yt', 'zt']:
            savestr = output_dirpath + '/orcasong_output/4dTo2d/' + proj + '/' + filename_output + '_' + proj + '.h5'
            pipe.attach(kp.io.HDF5Sink, filename=savestr, blob_keys=[proj, 'event_track'], complib=complib,
                        complevel=complevel, chunksize=chunksize, flush_frequency=flush_freq)


    if do3d:
        for proj in ['xyz', 'xyt', 'xzt', 'yzt', 'rzt']:
            savestr = output_dirpath + '/orcasong_output/4dTo3d/' + proj + '/' + filename_output + '_' + proj + '.h5'
            pipe.attach(kp.io.HDF5Sink, filename=savestr, blob_keys=[proj, 'event_track'], complib=complib,
                        complevel=complevel, chunksize=chunksize, flush_frequency=flush_freq)

    if do4d[0]:
        proj = 'xyzt' if not do4d[1] == 'channel_id' else 'xyzc'
        savestr = output_dirpath + '/orcasong_output/4dTo4d/' + proj + '/' + filename_output + '_' + proj + '.h5'
        pipe.attach(kp.io.HDF5Sink, filename=savestr, blob_keys=[proj, 'event_track'], complib=complib,
                    complevel=complevel, chunksize=chunksize, flush_frequency=flush_freq)

    # Execute Pipeline
    pipe.drain()

    if do2d_plots[0] is True:
        pdf_2d_plots.close()


def main():
    """
    Parses the input to the main make_nn_images function.
    """
    fname, detx_filepath, config_filepath = parse_input()
    config = load_config(config_filepath)
    check_user_input(fname, detx_filepath, config)
    make_nn_images(fname, detx_filepath, config)


if __name__ == '__main__':
    main()













#!/usr/bin/env python
# coding=utf-8
# Filename: utils.py

"""
Utility code for OrcaSong.
"""

import numpy as np
import km3pipe as kp


class EventSkipper(kp.Module):
    """
    KM3Pipe module that skips events based on some data_cuts.
    """
    def configure(self):
        self.data_cuts = self.require('data_cuts')

    def process(self, blob):
        continue_bool = skip_event(blob['event_track'], self.data_cuts)
        if continue_bool:
            return
        else:
            return blob


def skip_event(event_track, data_cuts):
    """
    Function that checks if an event should be skipped, depending on the data_cuts input.

    Parameters
    ----------
    event_track : ndarray(ndim=1)
        1D array containing important MC information of the event, only the energy of the event (pos_2) is used here.
    data_cuts : dict
        Dictionary that contains information about any possible cuts that should be applied.
        Supports the following cuts: 'triggered', 'energy_lower_limit', 'energy_upper_limit', 'throw_away_prob'.

    Returns
    -------
    continue_bool : bool
        boolean flag to specify, if this event should be skipped or not.

    """
    continue_bool = False

    if data_cuts['energy_lower_limit'] is not None:
        continue_bool = event_track.energy[0] < data_cuts['energy_lower_limit'] # True if E < lower limit

    if data_cuts['energy_upper_limit'] is not None and continue_bool == False:
        continue_bool = event_track.energy[0] > data_cuts['energy_upper_limit'] # True if E > upper limit

    if data_cuts['throw_away_prob'] is not None and continue_bool == False:
        throw_away_prob = data_cuts['throw_away_prob']
        throw_away = np.random.choice([False, True], p=[1 - throw_away_prob, throw_away_prob])
        if throw_away: continue_bool = True

    if data_cuts['custom_skip_function'] is not None:
        continue_bool = use_custom_skip_function(data_cuts['custom_skip_function'], event_track)

    return continue_bool


def use_custom_skip_function(skip_function, event_track):
    """
    User defined, custom skip functions.

    Parameters
    ----------
    skip_function : str
        String that specifies, which custom skip function should be used.
    event_track : ndarray(ndim=1)
        Structured numpy array containing the McTracks info of this event.

    Returns
    -------
    continue_bool : bool
        Bool which specifies, if this event should be skipped or not.

    """
    if skip_function == 'ts_e_flat':
        # cf. orcanet_contrib/utilities/get_func_for_flat_track_shower.py

        particle_type = event_track.particle_type[0]
        if np.abs(particle_type) not in [12, 14]:
            continue_bool = False

        else:
            prob_for_e_bin_arr_path = '/home/woody/capn/mppi033h/Code/OrcaSong/scraps/utilities/arr_fract_for_e_bins.npy'
            # Fraction of tracks compared to showers (n_muon-cc / (n_e-cc + n_e-nc))
            arr_fract_for_e_bins = np.load(prob_for_e_bin_arr_path) # 2d arr, one row e_bins, second row prob
            e_bins = arr_fract_for_e_bins[0, :]
            fract_e = arr_fract_for_e_bins[1, :]

            e_evt = event_track.energy[0]
            idx_e = (np.abs(e_bins - e_evt)).argmin()

            fract = fract_e[idx_e]

            if np.abs(particle_type) == 14: # for muon neutrinos
                if fract <= 1:
                    continue_bool = False
                else:
                    evt_survive_prob = 1/float(fract)
                    continue_bool = np.random.choice([False, True], p=[evt_survive_prob, 1 - evt_survive_prob])

            else:  # for elec neutrinos
                assert np.abs(particle_type) == 12

                if fract >= 1:
                    continue_bool = False
                else:
                    evt_survive_prob = fract
                    continue_bool = np.random.choice([False, True], p=[evt_survive_prob, 1 - evt_survive_prob])

        return continue_bool

    else:
        raise ValueError('The custom skip function "' + str(skip_function) + '" is not known.')


def get_file_particle_type(fname):
    """
    Returns a string that specifies the type of the particles that are contained in the input file.

    Parameters
    ----------
    fname : str
        Filename (full path!) of the input file.

    Returns
    -------
    file_particle_type : str
        String that specifies the type of particles that are contained in the file: ['undefined', 'muon', 'neutrino'].

    """
    event_pump = kp.io.hdf5.HDF5Pump(filename=fname, verbose=False) # TODO suppress print of hdf5pump

    if 'McTracks' not in event_pump[0]:
        file_particle_type = 'undefined'
    else:
        particle_type = event_pump[0]['McTracks'].type

        # if mupage file: first mc_track is an empty neutrino track, second track is the first muon
        if particle_type[0] == 0 and np.abs(particle_type[1]) == 13:
            file_particle_type = 'muon'

        # the first mc_track is the primary neutrino if the input file is containing neutrino events
        elif np.abs(particle_type[0]) in [12, 14, 16]:
            file_particle_type = 'neutrino'
        else:
            raise ValueError('The type of the particles in the "McTracks" folder, <', str(particle_type), '> is not known.')

    event_pump.close_file()

    return file_particle_type





from unittest import TestCase
import numpy as np
import orcasong.modules as modules
import km3pipe as kp


__author__ = 'Stefan Reck'


class TestModules(TestCase):
    def test_mc_info_maker(self):
        """ Test the mcinfo maker on some dummy data. """
        def extractor(blob):
            hits = blob["Hits"]
            return {"dom_id_0": hits.dom_id[0],
                    "time_2": hits.time[2]}

        in_blob = {
            "Hits": kp.Table({
                'dom_id': [2, 3, 3],
                'channel_id': [0, 1, 2],
                'time': [10.1, 11.2, 12.3]
            })
        }
        module = modules.McInfoMaker(
            extractor=extractor, store_as="test")
        out_blob = module.process(in_blob)

        self.assertSequenceEqual(list(out_blob.keys()), ["Hits", "test"])
        self.assertSequenceEqual(list(out_blob["test"].dtype.names),
                                 ('dom_id_0', 'time_2'))
        np.testing.assert_array_equal(out_blob["test"]["dom_id_0"],
                                      np.array([2, ], dtype="float64"))
        np.testing.assert_array_equal(out_blob["test"]["time_2"],
                                      np.array([12.3, ], dtype="float64"))

    def test_mc_info_maker_dtype(self):
        """ Test the mcinfo maker on some dummy data. """
        def extractor(blob):
            hits = blob["Hits"]
            return {"dom_id_0": hits.dom_id[0],
                    "time_2": hits.time[2]}

        in_blob = {
            "Hits": kp.Table({
                'dom_id': np.array([2, 3, 3], dtype="int8"),
                'time': np.array([10.1, 11.2, 12.3], dtype="float32"),
            })
        }
        module = modules.McInfoMaker(
            extractor=extractor, store_as="test", to_float64=False)
        out_blob = module.process(in_blob)

        np.testing.assert_array_equal(
            out_blob["test"]["dom_id_0"], np.array([2, ], dtype="int8"))
        np.testing.assert_array_equal(
            out_blob["test"]["time_2"], np.array([12.3, ], dtype="float32"))

    def test_event_skipper(self):
        def event_skipper(blob):
            # skip if true
            return blob["a"] == 42

        module = modules.EventSkipper(event_skipper=event_skipper)

        self.assertEqual(module.process({"a": 42}), None)
        self.assertEqual(module.process({"a": 25}), {"a": 25})


class TestTimePreproc(TestCase):
    def setUp(self):
        self.in_blob = {
            "Hits": kp.Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        self.in_blob_mc = {
            "Hits": kp.Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "McHits": kp.Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

    def test_time_preproc_t0(self):
        module = modules.TimePreproc(
            add_t0=True, center_time=False)

        target = {
            "Hits": kp.Table({
                'time': [1.1, 2.2, 3.3],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        out_blob = module.process(self.in_blob)

        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_equal(np.array(out_blob["Hits"]),
                                      np.array(target["Hits"]))

    def test_time_preproc_center(self):
        module = modules.TimePreproc(
            add_t0=False, center_time=True)

        target = {
            "Hits": kp.Table({
                'time': [-1., 0., 1.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        out_blob = module.process(self.in_blob)

        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_equal(np.array(out_blob["Hits"]),
                                      np.array(target["Hits"]))

    def test_time_preproc_t0_and_center(self):
        module = modules.TimePreproc(
            add_t0=True, center_time=True)

        target = {
            "Hits": kp.Table({
                'time': [-1.1, 0., 1.1],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        out_blob = module.process(self.in_blob)

        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["Hits"].view("<f8")),
            np.array(target["Hits"].view("<f8")))

    def test_time_preproc_mchits_t0_and_center(self):
        module = modules.TimePreproc(
            add_t0=True, center_time=True)

        target = {
            "Hits": kp.Table({
                'time': [-1.1, 0., 1.1],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "McHits": kp.Table({
                'time': [-1.2, -0.2, 0.8],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
        }
        out_blob = module.process(self.in_blob_mc)

        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["McHits"].view("<f8")),
            np.array(target["McHits"].view("<f8")))


class TestPointMaker(TestCase):
    def setUp(self):
        self.input_blob_1 = {
            "Hits": kp.Table({
                "x": [4, 5, 6],
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],}),
            "EventInfo": kp.Table({
                "pad": 1.
            })
        }

    def test_default_settings(self):
        pm = modules.PointMaker(
            max_n_hits=4)
        result = pm.process(self.input_blob_1)["samples"]
        self.assertTupleEqual(
            pm.finish()["hit_infos"], ("t0", "time", "x", "is_valid"))
        target = np.array(
            [[[0.1, 1, 4, 1],
              [0.2, 2, 5, 1],
              [0.3, 3, 6, 1],
              [0,   0, 0, 0]]], dtype="float32")
        np.testing.assert_array_equal(result, target)

    def test_input_blob_1(self):
        pm = modules.PointMaker(
            max_n_hits=4,
            hit_infos=("x", "time"),
            time_window=None,
            dset_n_hits=None,
        )
        result = pm.process(self.input_blob_1)["samples"]
        self.assertTupleEqual(
            pm.finish()["hit_infos"], ("x", "time", "is_valid"))
        target = np.array(
            [[[4, 1, 1],
              [5, 2, 1],
              [6, 3, 1],
              [0, 0, 0]]], dtype="float32")
        np.testing.assert_array_equal(result, target)

    def test_input_blob_1_max_n_hits(self):
        input_blob_long = {
            "Hits": kp.Table({
                "x": np.random.rand(1000).astype("float32"),
        })}
        result = modules.PointMaker(
            max_n_hits=10,
            hit_infos=("x",),
            time_window=None,
            dset_n_hits=None,
        ).process(input_blob_long)["samples"]

        self.assertSequenceEqual(result.shape, (1, 10, 2))
        self.assertTrue(all(
            np.isin(result[0, :, 0], input_blob_long["Hits"]["x"])))

    def test_input_blob_time_window(self):
        result = modules.PointMaker(
            max_n_hits=4,
            hit_infos=("x", "time"),
            time_window=[1, 2],
            dset_n_hits=None,
        ).process(self.input_blob_1)["samples"]
        target = np.array(
            [[[4, 1, 1],
              [5, 2, 1],
              [0, 0, 0],
              [0, 0, 0]]], dtype="float32")
        np.testing.assert_array_equal(result, target)

    def test_input_blob_time_window_nhits(self):
        result = modules.PointMaker(
            max_n_hits=4,
            hit_infos=("x", "time"),
            time_window=[1, 2],
            dset_n_hits="EventInfo",
        ).process(self.input_blob_1)["EventInfo"]
        print(result)
        self.assertEqual(result["n_hits_intime"], 2)


class TestImageMaker(TestCase):
    def test_2d_xt_binning(self):
        # (3 x 2) x-t binning
        bin_edges_list = [
            ["x", [3.5, 4.5, 5.5, 6.5]],
            ["time", [0.5, 2, 3.5]]
        ]

        module = modules.ImageMaker(bin_edges_list=bin_edges_list)
        in_blob = {
            "Hits": kp.Table({
                "x": [4, 5, 6],
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        target = {
            "Hits": kp.Table({
                "x": [4, 5, 6],
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "samples": np.array([[
                [1, 0],
                [0, 1],
                [0, 1],
            ]])
        }

        out_blob = module.process(in_blob)
        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["Hits"].view("<f8")),
            np.array(target["Hits"].view("<f8")))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["samples"]),
            np.array(target["samples"]))

    def test_unknown_field(self):
        # (3 x 2) x-t binning
        bin_edges_list = [
            ["aggg", [3.5, 4.5, 5.5, 6.5]],
            ["time", [0.5, 2, 3.5]]
        ]

        module = modules.ImageMaker(
            bin_edges_list=bin_edges_list)
        in_blob = {
            "Hits": kp.Table({
                "x": [4, 5, 6],
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        with self.assertRaises(ValueError):
            module.process(in_blob)

    def test_1d_binning(self):
        # (1, ) t binning
        bin_edges_list = [
            ["time", [2.5, 3.5]]
        ]

        module = modules.ImageMaker(bin_edges_list=bin_edges_list)
        in_blob = {
            "Hits": kp.Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        target = {
            "Hits": kp.Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "samples": np.array([
                [1, ],
            ])
        }

        out_blob = module.process(in_blob)
        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["Hits"].view("<f8")),
            np.array(target["Hits"].view("<f8")))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["samples"]),
            np.array(target["samples"]))

    def test_1d_binning_no_hits(self):
        # (1, ) t binning
        bin_edges_list = [
            ["time", [3.5, 4.5]]
        ]

        module = modules.ImageMaker(bin_edges_list=bin_edges_list)
        in_blob = {
            "Hits": kp.Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            })
        }

        target = {
            "Hits": kp.Table({
                'time': [1., 2., 3.],
                "t0": [0.1, 0.2, 0.3],
                "triggered": [0, 1, 1],
            }),
            "samples": np.array([
                [0, ],
            ])
        }

        out_blob = module.process(in_blob)
        self.assertSetEqual(set(out_blob.keys()), set(target.keys()))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["Hits"].view("<f8")),
            np.array(target["Hits"].view("<f8")))
        np.testing.assert_array_almost_equal(
            np.array(out_blob["samples"]),
            np.array(target["samples"]))


class TestBinningStatsMaker(TestCase):
    def test_it(self):
        # (3 x 2) x-t binning
        bin_edges_list = [
            ["x", [3.5, 4.5, 5.5, 6.5]],
            ["time", [0.5, 2, 3.5]],
            ["z", [1, 4]]
        ]

        in_blob = {
            "Hits": kp.Table({
                "x": [4, 5, 6, 6],
                'time': [1., 2., 3., 50],
                "z": [0, 3, 4, 5],

                "t0": [0.1, 0.2, 0.3, 0.4],
                "triggered": [0, 1, 1, 1],
            })
        }

        target = {
            'x': {
                'hist': np.array([0., 0., 0., 1., 0., 1.]),
                'hist_bin_edges': np.array([3.5, 4., 4.5, 5., 5.5, 6., 6.5]),
                'bin_edges': [3.5, 4.5, 5.5, 6.5],
                'cut_off': np.array([0., 0.])
            },
            'time': {
                'hist': np.array([0., 2.]),
                'hist_bin_edges': [0.5, 2, 3.5],
                'bin_edges': [0.5, 2, 3.5],
                'cut_off': np.array([0., 1.])
            },
            'z': {
                'hist': np.array([0., 2.]),
                'hist_bin_edges': np.array([1., 2.5, 4.]),
                'bin_edges': [1, 4],
                'cut_off': np.array([1., 1.])
            }
        }

        module = modules.BinningStatsMaker(
            bin_edges_list=bin_edges_list, res_increase=2)
        module.process(in_blob)
        output = module.finish()
        check_dicts_n_ray(output, target)


def check_dicts_n_ray(a, b):
    """ Check if dicts with dicts with ndarrays are equal. """
    if set(a.keys()) != set(b.keys()):
        raise KeyError("{} != {}".format(a.keys(), b.keys()))
    for key in a.keys():
        if set(a[key].keys()) != set(b[key].keys()):
            raise KeyError("{} != {}".format(a[key].keys(), b[key].keys()))
        for skey in a[key].keys():
            np.testing.assert_array_almost_equal(a[key][skey], b[key][skey])



"""
Distribute filepaths among lists, so that they can be concatenated to a
given number of files, and with a train/val split.

Main function: get_split
    The output will be a list with dicts containing a name for the concatenated
    file, and the paths of the individual files that will be concatenated.
    This can be handed over to concatenate_h5 in orcasong.

Example:
    For a folder with 4 files named file_0.h5 to file_3.h5 and a 50:50 split
    between train/val:

    get_split(folder,
              outfile_basestr='test',
              n_train_files=1,
              n_val_files=1,
              train_frac=0.5)

    [
     {
      'file_list': array(['file_2.h5', file_1.h5]),
      'output_filepath': 'test_train_0.h5',
      'is_train': True,
     },
     {
      'file_list': array(['file_4.h5', 'file_0.h5']),
      'output_filepath': 'test_val_0.h5',
      'is_train': False,
     },
    ]

"""
import os
import numpy as np

__author__ = 'Stefan Reck'


def get_files(folder):
    """
    Get pathes of all h5 files in given folder.
    """
    infiles = os.listdir(folder)
    infiles.sort()

    infile_paths = []
    for infile in infiles:
        if infile.endswith(".h5"):
            infile_paths.append(os.path.join(folder, infile))

    return np.array(infile_paths)


def split_path_list(files, train_frac, n_train_files, n_val_files):
    """
    Get train and val files split according to given fraction, and
    distributed over given number of files.

    Parameters
    ----------
    files : List
        The files.
    train_frac : float
        The fraction of files.
    n_train_files : int
        Total number of resulting train files.
    n_val_files : int
        Total number of resulting val files.

    Returns
    -------
    job_files_train : ndarray
        The train files. They are chosen randomly from the files list.
        The total number of files is the given fraction of the input files.
    job_files_val : ndarray
        The val files, similar to the train files.

    """
    if n_train_files < 1 or n_val_files < 1:
        raise ValueError("Need at least 1 file for train and val.")

    order = np.arange(len(files))
    np.random.shuffle(order)

    len_train_files = int(len(files) * train_frac)
    train_files = files[order[:len_train_files]]
    val_files = files[order[len_train_files:]]

    job_files_train = np.array_split(train_files, n_train_files)
    job_files_val = np.array_split(val_files, n_val_files)

    for fs in job_files_train:
        if len(fs) == 0:
            raise ValueError("No files for an output train file!")

    for fs in job_files_val:
        if len(fs) == 0:
            raise ValueError("No files for an output val file!")

    return job_files_train, job_files_val


def get_split(folder, outfile_basestr, n_train_files=1, n_val_files=1,
              train_frac=0.8, seed=64):
    """
    Distribute filepaths among lists (see doc at top of file).

    The files in each train or val file will be drawn randomly from the
    available files. Each train or val files will be created by its own
    seperately submitted job.

    Parameters
    ----------
    folder : str
        Containing the files to concat.
    n_train_files : int
        Total number of resulting train files.
    n_val_files : int
        Total number of resulting val files.
    outfile_basestr : str
        Path and a base for the name. "train"/"val" and a file number will
        get automatically attached to the name.
    train_frac : float
        The fraction of files in the train set.
    seed : int, optional
        Seed for the random stuff.

    Returns
    -------
    jobs : dict
        Contains the arguments for the concatenate script.

    """
    if seed is not None:
        np.random.seed(seed)

    files = get_files(folder)
    job_files_train, job_files_val = split_path_list(files, train_frac, n_train_files, n_val_files)

    jobs = []

    for i, job_files in enumerate(job_files_train):
        output_filepath = "{}_train_{}.h5".format(outfile_basestr, i)
        job_dict = {
            "file_list": job_files,
            "output_filepath": output_filepath,
            "is_train": True,
        }
        jobs.append(job_dict)

    for i, job_files in enumerate(job_files_val):
        output_filepath = "{}_val_{}.h5".format(outfile_basestr, i)
        job_dict = {
            "file_list": job_files,
            "output_filepath": output_filepath,
            "is_train": False,
        }
        jobs.append(job_dict)

    return jobs



#!/usr/bin/env python
# Filename: __version__.py
# pylint: disable=C0103
"""
Pep 386 compliant version info.

    (major, minor, micro, alpha/beta/rc/final, #)
    (1, 1, 2, 'alpha', 0) => "1.1.2.dev"
    (1, 2, 0, 'beta', 2) => "1.2b2"

"""
from __future__ import absolute_import, print_function, division
from os.path import dirname, realpath, join
from setuptools_scm import get_version

version = 'unknown version'

try:
    version = get_version(root='..', relative_to=__file__)
except LookupError:
    try:
        with open(join(realpath(dirname(__file__)), "version.txt"),
                  'r') as fobj:
            version = fobj.read()
    except IOError:
        pass



#!/usr/bin/env python
from setuptools import setup, find_packages
# from pkg_resources import get_distribution, DistributionNotFound

with open('requirements.txt') as fobj:
    requirements = [l.strip() for l in fobj.readlines()]

setup(
    name='orcasong',
    description='Makes images for a NN based on the hit information of neutrino '
                'events in the neutrino telescope KM3NeT',
    url='https://git.km3net.de/ml/OrcaSong',
    author='Stefan Reck, Michael Moser',
    author_email='stefan.reck@fau.de, mmoser@km3net.de, michael.m.moser@fau.de',
    license='AGPL',
    install_requires=requirements,
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python', ],

    setup_requires=['setuptools_scm'],
    use_scm_version={'write_to': 'orcasong/version.txt',
                     'tag_regex': r'^(?P<prefix>v)?(?P<version>[^\+]+)(?P<suffix>.*)?$', },

    entry_points={'console_scripts': [
        'concatenate=orcasong.tools.concatenate:main',
        'h5shuffle=orcasong.tools.postproc:h5shuffle',
        'plot_binstats=orcasong.plotting.plot_binstats:main',
        'make_nn_images=legacy.make_nn_images:main',
        'make_dsplit=orcasong_contrib.data_tools.make_data_split.make_data_split:main']}

)

__author__ = 'Stefan Reck, Michael Moser'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Concatenates .h5 files. Works only for files where each dataset has the same number of rows."""

import h5py
import numpy as np
import math
from argparse import ArgumentParser, RawTextHelpFormatter
import sys
# from memory_profiler import profile # for memory profiling, call with @profile; myfunc()

__author__ = 'Michael Moser'
__license__ = 'AGPL'
__version__ = '1.0'
__email__ = 'michael.m.moser@fau.de'
__status__ = 'Production'


def parse_input():
    """
    Parses the user input in order to return the most important information:

    1) list of files that should be concatenated
    2) the filepath of the output .h5 file
    3) use custom chunksize or not.

    Returns
    -------
    file_list : list
        List that contains all filepaths of the input files.
    output_filepath : str
        String that specifies the filepath (path+name) of the output .h5 file.
    chunksize : None/int
        Specifies the chunksize for axis_0 in the concatenated output files.
        If None, the chunksize is read from the first input file.
        Else, a custom chunksize will be used.
    complib : None/str
        Specifies the compression library that should be used for saving the concatenated output files.
        If None, the compression library is read from the first input file.
        Else, a custom compression library will be used.
        Currently available: 'gzip', or 'lzf'.
    complevel : None/int
        Specifies the compression level that should be used for saving the concatenated output files.
        A compression level is only available for gzip compression, not lzf!
        If None, the compression level is read from the first input file.
        Else, a custom compression level will be used.

    """
    parser = ArgumentParser(description='E.g. < python concatenate_h5.py file_1 file_2 /path/to/output.h5 > or '
                                        '< python concatenate_h5.py --list filepaths.txt /path/to/output.h5 >.\n'
                                        'Concatenates arrays stored in .h5 files for either multiple direct .h5 inputs or a .txt file of .h5 files (--list option).\n'
                                        'Outputs a new .h5 file with the concatenated arrays. This output is chunked!\n'
                                        'Careful: The folders of one file need to have the same number of rows (axis_0)!\n'
                                        'Make a .txt file with < find /path/to/files -name "file_x-*.h5" | sort --version-sort > listname.list >\n'
                                        'Chunksize: By default, the chunksize is set to the chunksize of the first inputfile!',
                            formatter_class=RawTextHelpFormatter)

    parser.add_argument('files', metavar='file', type=str, nargs='*', help = 'a file that should be concatenated, minimum of two.')
    parser.add_argument('output_filepath', metavar='output_filepath', type=str, nargs=1, help='filepath and name of the output .h5 file')
    parser.add_argument('-l', '--list', dest='list_file', type=str,
                        help='filepath of a .list file that contains all .h5 files that should be concatenated')
    parser.add_argument('--chunksize', dest='chunksize', type=int,
                        help='Specify a chunksize value in order to use chunked storage for the concatenated .h5 file.'
                             ' Otherwise, it will be read from the first input file..')
    parser.add_argument('--complib', dest='complib', type=str,
                        help='Specify a filter that should be used for compression. Either "gzip" or "lzf". '
                             'Otherwise, the filter will be read from the first input file.')
    parser.add_argument('--complevel', dest='complevel', type=int,
                        help='Specify a compression filter strength that should be used for the compression. '
                             'Otherwise, the filter will be read from the first input file. '
                             'Can range from 0 to 9. Has no effect on "lzf" compression.')

    if len(sys.argv) == 1:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args()

    if args.list_file:
        file_list = []
        for line in open(args.list_file):
            filepath = line.rstrip('\n')
            if filepath != "":
                file_list.append(filepath)
    else:
        file_list = []
        for filepath in args.files:
            file_list.append(filepath)

    output_filepath = args.output_filepath[0]

    chunksize = None
    if args.chunksize:
        chunksize = args.chunksize
        print('You chose chunksize = ' + str(chunksize))

    complib = None
    if args.complib:
        complib = args.complib
        print('You chose complib = ' + complib)

    complevel = None
    if args.complevel:
        complevel = args.complevel
        print('You chose complevel = ' + str(complevel))

    return file_list, output_filepath, chunksize, complib, complevel


def get_cum_number_of_rows(file_list):
    """
    Returns the cumulative number of rows (axis_0) in a list based on the
    specified input .h5 files (i.e. [0,100,200,300,...] if each file has
    100 rows).

    Parameters
    ----------
    file_list : list
        List that contains all filepaths of the input files.

    Returns
    -------
    cum_number_of_rows_list : list
        List that contains the cumulative number of rows
        (i.e. [0,100,200,300,...] if each file has 100 rows).

    """
    total_number_of_rows = 0
    cum_number_of_rows_list = [0]
    number_of_rows_list = []  # used for approximating the chunksize

    # Get total number of rows for the files in the list, faster than resizing
    # the dataset in each iteration of the file loop in concatenate_h5_files()

    for file_name in file_list:
        with h5py.File(file_name, 'r') as f:
            # get number of rows from the first folder of the file
            # -> each folder needs to have the same number of rows
            f_keys = list(f.keys())
            f_keys_stripped = strip_keys(f_keys)

            total_number_of_rows += f[f_keys_stripped[0]].shape[0]
            cum_number_of_rows_list.append(total_number_of_rows)
            number_of_rows_list.append(f[f_keys_stripped[0]].shape[0])

    return cum_number_of_rows_list


def strip_keys(f_keys):
    """
    Remove pytables folders starting with '_i_', because the shape
    of its first axis does not correspond to the number of events
    in the file. All other folders normally have an axis_0 shape
    that is equal to the number of events in the file.
    Also remove bin_stats.
    """
    f_keys_stripped = []
    for x in f_keys:
        if is_folder_ignored(x):
            continue
        f_keys_stripped.append(x)
    return f_keys_stripped


def is_folder_ignored(folder_name):
    """
    Defines pytable folders which should be ignored during concat.
    """
    if '_i_' in folder_name or "bin_stats" in folder_name:
        return True
    else:
        return False


def get_f_compression_and_chunking(filepath):
    """
    Function that gets the used compression library, the compression level (if applicable)
    and the chunksize of axis_0 of the first dataset of the file.

    Parameters
    ----------
    filepath : str
        Filepath of a .hdf5 file.

    Returns
    -------
    compression : str
        The compression library that has been identified in the input file. E.g. 'gzip', or 'lzf'.
    complevel : int
        The compression level that has been identified in the input file.
    chunksize : None/int
        The chunksize of axis_0 that has been indentified in the input file.

    """
    f = h5py.File(filepath, 'r')

    # remove any keys to pytables folders that may be in the file
    f_keys = list(f.keys())
    f_keys_stripped = strip_keys(f_keys)

    compression = f[f_keys_stripped[0]].compression  # compression filter
    compression_opts = f[f_keys_stripped[0]].compression_opts  # filter strength
    chunksize = f[f_keys_stripped[0]].chunks[0]  # chunksize along axis_0 of the dataset

    return compression, compression_opts, chunksize


def concatenate_h5_files(output_filepath, file_list,
                         chunksize=None, complib=None, complevel=None):
    """
    Function that concatenates hdf5 files based on an output_filepath and a file_list of input files.

    If the files contain group_info and x_indices folders (if the input files are coming from km3pipe output),
    the group-id / the index of the x_indices is fixed in order to not get duplicates of group-ids / x-indices.

    Parameters
    ----------
    output_filepath : str
        String that specifies the filepath (path+name) of the output .h5 file.
    file_list : list
        List that contains all filepaths of the input files.
    chunksize : None/int
        Specifies the chunksize for axis_0 in the concatenated output files.
        If None, the chunksize is read from the first input file.
        Else, a custom chunksize will be used.
    complib : None/str
        Specifies the compression library that should be used for saving the concatenated output files.
        If None, the compression library is read from the first input file.
        Else, a custom compression library will be used.
        Currently available: 'gzip', or 'lzf'.
    complevel : None/int
        Specifies the compression level that should be used for saving the concatenated output files.
        A compression level is only available for gzip compression, not lzf!
        If None, the compression level is read from the first input file.
        Else, a custom compression level will be used.

    """
    cum_rows_list = get_cum_number_of_rows(file_list)
    complib_f, complevel_f, chunksize_f = get_f_compression_and_chunking(file_list[0])

    chunksize = chunksize_f if chunksize is None else chunksize
    complib = complib_f if complib is None else complib
    complevel = complevel_f if complevel is None else complevel

    if complib == 'lzf':
        complevel = None

    file_output = h5py.File(output_filepath, 'w')

    for n, input_file_name in enumerate(file_list):
        print('Processing file ' + file_list[n])
        input_file = h5py.File(input_file_name, 'r')

        # create metadata
        if 'format_version' in list(input_file.attrs.keys()) and n == 0:
            file_output.attrs['format_version'] = input_file.attrs['format_version']

        for folder_name in input_file:
            if is_folder_ignored(folder_name):
                # we ignore datasets that have been created by pytables, don't need them anymore
                continue

            if n > 0 and folder_name in ['event_info', 'group_info', 'x_indices', 'y']:
                folder_data = input_file[folder_name][()]
                # we need to add the current number of the group_id / index in the file_output
                # to the group_ids / indices of the file that is to be appended
                column_name = 'group_id' if folder_name in ['event_info', 'group_info', 'y'] else 'index'
                # add 1 because the group_ids / indices start with 0
                folder_data[column_name] += np.amax(file_output[folder_name][column_name]) + 1

            else:
                folder_data = input_file[folder_name]

            print('Shape and dtype of dataset ' + folder_name + ': ' + str(folder_data.shape) + ' ; ' + str(folder_data.dtype))

            if n == 0:
                # first file; create the dummy dataset with no max shape
                maxshape = (None,) + folder_data.shape[1:]  # change shape of axis zero to None
                chunks = (chunksize,) + folder_data.shape[1:]

                output_dataset = file_output.create_dataset(folder_name, data=folder_data, maxshape=maxshape, chunks=chunks,
                                                            compression=complib, compression_opts=complevel)

                output_dataset.resize(cum_rows_list[-1], axis=0)

            else:
                file_output[folder_name][cum_rows_list[n]:cum_rows_list[n+1]] = folder_data

        file_output.flush()

    print('Output information:')
    print('-------------------')
    print('The output file contains the following datasets:')
    for folder_name in file_output:
        print('Dataset ' + folder_name + ' with the following shape, dtype and chunks (first argument'
              ' is the chunksize in axis_0): \n' + str(file_output[folder_name].shape) + ' ; ' +
              str(file_output[folder_name].dtype) + ' ; ' + str(file_output[folder_name].chunks))

    file_list_ascii = [n.encode("ascii", "ignore") for n in file_list]
    file_output.create_dataset("used_files", data=file_list_ascii)

    file_output.close()


def main():
    """
    Main code. Concatenates .h5 files with multiple datasets, where each dataset in one file needs to have the same number of rows (axis_0).

    Gets user input with aid of the parse_input() function. By default, the chunksize for the output .h5 file is automatically computed.
    based on the average number of rows per file, in order to eliminate padding (wastes disk space).
    For faster I/O, the chunksize should be set by the user depending on the use case.
    In deep learning applications for example, the chunksize should be equal to the batch size that is used later on for reading the data.
    """
    file_list, output_filepath, chunksize, complib, complevel = parse_input()
    concatenate_h5_files(output_filepath, file_list, chunksize, complib, complevel)


if __name__ == '__main__':
    main()
